from typing import Any, Dict, List
from aiida import orm
from aiida_workgraph.orm.serializer import general_serializer
from aiida_workgraph.task import Task


class PythonJob(Task):
    """PythonJob Task."""

    identifier = "workgraph.pythonjob"

    function_kwargs: List = None

    def update_from_dict(self, data: Dict[str, Any], **kwargs) -> "PythonJob":
        """Overwrite the update_from_dict method to handle the PythonJob data."""
        self.function_kwargs = data.get("function_kwargs", [])
        self.deserialize_pythonjob_data(data)
        super().update_from_dict(data)

    def to_dict(self) -> Dict[str, Any]:
        data = super().to_dict()
        data["function_kwargs"] = self.function_kwargs
        return data

    @classmethod
    def serialize_pythonjob_data(cls, tdata: Dict[str, Any]):
        """Serialize the properties for PythonJob."""

        input_kwargs = tdata.get("function_kwargs", [])
        for name in input_kwargs:
            tdata["inputs"][name]["property"]["value"] = cls.serialize_socket_data(
                tdata["inputs"][name]
            )

    @classmethod
    def deserialize_pythonjob_data(cls, tdata: Dict[str, Any]) -> None:
        """
        Process the task data dictionary for a PythonJob.
        It load the orignal Python data from the AiiDA Data node for the
        args and kwargs of the function.

        Args:
            tdata (Dict[str, Any]): The input data dictionary.

        Returns:
            Dict[str, Any]: The processed data dictionary.
        """
        input_kwargs = tdata.get("function_kwargs", [])

        for name in input_kwargs:
            if name in tdata["inputs"]:
                tdata["inputs"][name]["property"][
                    "value"
                ] = cls.deserialize_socket_data(tdata["inputs"][name])

    @classmethod
    def serialize_socket_data(cls, data: Dict[str, Any]) -> Any:
        if data.get("identifier", "Any").upper() == "WORKGRAPH.NAMESPACE":
            if data["property"]["value"] is None:
                return None
            if isinstance(data["property"]["value"], dict):
                serialized_result = {}
                for key, value in data["property"]["value"].items():
                    serialized_result[key] = general_serializer(value)
                return serialized_result
            else:
                raise ValueError("Namespace socket should be a dictionary.")
        else:
            if isinstance(data["property"]["value"], orm.Data):
                return data["property"]["value"]
            return general_serializer(data["property"]["value"])

    @classmethod
    def deserialize_socket_data(cls, data: Dict[str, Any]) -> Any:
        if data.get("identifier", "Any").upper() == "WORKGRAPH.NAMESPACE":
            if isinstance(data["property"]["value"], dict):
                deserialized_result = {}
                for key, value in data["property"]["value"].items():
                    if isinstance(value, orm.Data):
                        deserialized_result[key] = value.value
                    else:
                        deserialized_result[key] = value
                return deserialized_result
            else:
                raise ValueError("Namespace socket should be a dictionary.")
        else:
            if isinstance(data["property"]["value"], orm.Data):
                return data["property"]["value"].value
            return data["property"]["value"]



"""`Data` sub class to represent any data using pickle."""

from aiida import orm
import sys
import cloudpickle
from pickle import UnpicklingError


class Dict(orm.Dict):
    @property
    def value(self):
        return self.get_dict()


class List(orm.List):
    @property
    def value(self):
        return self.get_list()


class GeneralData(orm.Data):
    """Data to represent a pickled value using cloudpickle."""

    FILENAME = "value.pkl"  # Class attribute to store the filename

    def __init__(self, value=None, **kwargs):
        """Initialize a `GeneralData` node instance.

        :param value: raw Python value to initialize the `GeneralData` node from.
        """
        super().__init__(**kwargs)
        self.set_value(value)

    def __str__(self):
        return f"{super().__str__()} : {self.get_value()}"

    @property
    def value(self):
        """Return the contents of this node.

        :return: The unpickled value.
        """
        return self.get_value()

    @value.setter
    def value(self, value):
        self.set_value(value)

    def get_value(self):
        """Return the contents of this node, unpickling the stored value.

        :return: The unpickled value.
        """
        return self._get_value_from_file()

    def _get_value_from_file(self):
        """Read the pickled value from file and return it."""
        try:
            with self.base.repository.open(self.FILENAME, mode="rb") as f:
                return cloudpickle.loads(f.read())  # Deserialize the value
        except (UnpicklingError, ValueError) as e:
            raise ImportError(
                "Failed to load the pickled value. This may be due to an incompatible pickle protocol. "
                "Please ensure that the correct environment and cloudpickle version are being used."
            ) from e
        except ModuleNotFoundError as e:
            raise ImportError(
                "Failed to load the pickled value. This may be due to a missing module. "
                "Please ensure that the correct environment and cloudpickle version are being used."
            ) from e

    def set_value(self, value):
        """Set the contents of this node by pickling the provided value.

        :param value: The Python value to pickle and store.
        """
        # Serialize the value and store it
        serialized_value = cloudpickle.dumps(value)
        self.base.repository.put_object_from_bytes(serialized_value, self.FILENAME)

        # Store relevant metadata
        python_version = f"{sys.version_info.major}.{sys.version_info.minor}"
        self.base.attributes.set("python_version", python_version)
        self.base.attributes.set("serializer_module", cloudpickle.__name__)
        self.base.attributes.set("serializer_version", cloudpickle.__version__)
        self.base.attributes.set("pickle_protocol", cloudpickle.DEFAULT_PROTOCOL)



def svg_to_html(svg_xml: str, width: str = "100%", height: str = "100%") -> str:
    """
    Converts an SVG XML string into an HTML string with embedded SVG,
    scaled to the specified width and height using CSS and includes functionality
    for panning and zooming the SVG based on the mouse point.
    """
    html_template = f"""
    <!DOCTYPE html>
    <html lang="en">
    <head>
        <meta charset="UTF-8">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
        <title>Interactive SVG Viewer</title>
        <style>
            svg {{
                width: {width};
                height: {height};
                cursor: grab;
            }}
            svg:active {{
                cursor: grabbing;
            }}
            #fullscreen-btn {{
                padding: 5px 10px;
            }}
            #fullscreen-btn:hover {{
                background-color: #0056b3;
                color: white;
            }}
        </style>
    </head>
    <body>
        <button id="fullscreen-btn">Fullscreen</button>
        {svg_xml}
        <script>
            const svgElement = document.querySelector('svg');
            let isPanning = false;
            let startPoint = {{x: 0, y: 0}};
            let translate = {{x: 0, y: 0}};
            let scale = 1;

            // Enable panning
            svgElement.addEventListener('mousedown', (e) => {{
                isPanning = true;
                startPoint = {{x: e.clientX - translate.x, y: e.clientY - translate.y}};
                svgElement.style.cursor = 'grabbing';
            }});

            window.addEventListener('mouseup', () => {{
                isPanning = false;
                svgElement.style.cursor = 'grab';
            }});

            window.addEventListener('mousemove', (e) => {{
                if (isPanning) {{
                    translate.x = e.clientX - startPoint.x;
                    translate.y = e.clientY - startPoint.y;
                    updateTransform();
                }}
            }});

            // Enable zooming
            svgElement.addEventListener('wheel', (e) => {{
                e.preventDefault();
                const mouseX = e.clientX;
                const mouseY = e.clientY;
                const scaleFactor = 1.03;
                if (e.deltaY < 0) {{
                    // Zoom in
                    scale *= scaleFactor;
                }} else {{
                    // Zoom out
                    scale /= scaleFactor;
                }}
                translate.x -= mouseX * (scaleFactor - 1) * (e.deltaY < 0 ? 1 : -1);
                translate.y -= mouseY * (scaleFactor - 1) * (e.deltaY < 0 ? 1 : -1);
                updateTransform();
            }});

            function updateTransform() {{
                svgElement.style.transform = `translate(${{translate.x}}px, ${{translate.y}}px) scale(${{scale}})`;
            }}

            // Fullscreen toggle
            const fullscreenButton = document.getElementById('fullscreen-btn');
            fullscreenButton.addEventListener('click', () => {{
                if (!document.fullscreenElement) {{
                    svgElement.requestFullscreen().catch(err => {{
                        alert(`Error attempting to enable full-screen mode: ${{err.message}} (${{err.name}})`);
                    }});
                }} else {{
                    document.exitFullscreen();
                }}
            }});
        </script>
    </body>
    </html>
    """
    return html_template



from aiida_workgraph.orm.serializer import serialize_to_aiida_nodes
from aiida import orm
from aiida.common.extendeddicts import AttributeDict


def prepare_for_workgraph_task(task: dict, kwargs: dict) -> tuple:
    """Prepare the inputs for WorkGraph task"""
    from aiida_workgraph.utils import merge_properties, serialize_properties
    from aiida.orm.utils.serialize import deserialize_unsafe

    wgdata = deserialize_unsafe(task["executor"]["wgdata"])
    wgdata["name"] = task["name"]
    wgdata["metadata"]["group_outputs"] = task["metadata"]["group_outputs"]
    # update the workgraph data by kwargs
    for task_name, data in kwargs.items():
        # because kwargs is updated using update_nested_dict_with_special_keys
        # which means the data is grouped by the task name
        for socket_name, value in data.items():
            wgdata["tasks"][task_name]["inputs"][socket_name]["property"][
                "value"
            ] = value
    # merge the properties
    merge_properties(wgdata)
    serialize_properties(wgdata)
    metadata = {"call_link_label": task["name"]}
    inputs = {"wg": wgdata, "metadata": metadata}
    return inputs, wgdata


def prepare_for_python_task(task: dict, kwargs: dict, var_kwargs: dict) -> dict:
    """Prepare the inputs for PythonJob"""
    from aiida_workgraph.utils import get_or_create_code
    import os

    # get the names kwargs for the PythonJob, which are the inputs before _wait
    function_kwargs = kwargs.pop("function_kwargs", {})
    # TODO better way to find the function_kwargs
    input_names = [
        name
        for name, _ in sorted(
            ((name, input["list_index"]) for name, input in task["inputs"].items()),
            key=lambda x: x[1],
        )
    ]
    for name in input_names:
        if name == "_wait":
            break
        function_kwargs[name] = kwargs.pop(name, None)
    # if the var_kwargs is not None, we need to pop the var_kwargs from the kwargs
    # then update the function_kwargs if var_kwargs is not None
    if task["var_kwargs"] is not None:
        function_kwargs.pop(task["var_kwargs"], None)
        if var_kwargs:
            # var_kwargs can be AttributeDict if it get data from the previous task output
            if isinstance(var_kwargs, (dict, AttributeDict)):
                function_kwargs.update(var_kwargs)
            # otherwise, it should be a Data node
            elif isinstance(var_kwargs, orm.Data):
                function_kwargs.update(var_kwargs.value)
            else:
                raise ValueError(f"Invalid var_kwargs type: {type(var_kwargs)}")
    # setup code
    code = kwargs.pop("code", None)
    computer = kwargs.pop("computer", None)
    code_label = kwargs.pop("code_label", None)
    code_path = kwargs.pop("code_path", None)
    prepend_text = kwargs.pop("prepend_text", None)
    upload_files = kwargs.pop("upload_files", {})
    new_upload_files = {}
    # change the string in the upload files to SingleFileData, or FolderData
    for key, source in upload_files.items():
        # only alphanumeric and underscores are allowed in the key
        # replace all "." with "_dot_"
        new_key = key.replace(".", "_dot_")
        if isinstance(source, str):
            if os.path.isfile(source):
                new_upload_files[new_key] = orm.SinglefileData(file=source)
            elif os.path.isdir(source):
                new_upload_files[new_key] = orm.FolderData(tree=source)
        elif isinstance(source, (orm.SinglefileData, orm.FolderData)):
            new_upload_files[new_key] = source
        else:
            raise ValueError(f"Invalid upload file type: {type(source)}, {source}")
    #
    if code is None:
        code = get_or_create_code(
            computer=computer if computer else "localhost",
            code_label=code_label if code_label else "python3",
            code_path=code_path if code_path else None,
            prepend_text=prepend_text if prepend_text else None,
        )
    metadata = kwargs.pop("metadata", {})
    metadata.update({"call_link_label": task["name"]})
    # get the source code of the function
    function_name = task["executor"]["name"]
    if task["executor"].get("is_pickle", False):
        function_source_code = (
            task["executor"]["import_statements"]
            + "\n"
            + task["executor"]["source_code_without_decorator"]
        )
    else:
        function_source_code = (
            f"from {task['executor']['module']} import {function_name}"
        )

    # outputs
    function_outputs = [
        output
        for output, _ in sorted(
            ((output, output["list_index"]) for output in task["outputs"].values()),
            key=lambda x: x[1],
        )
    ]
    # serialize the kwargs into AiiDA Data
    function_kwargs = serialize_to_aiida_nodes(function_kwargs)
    # transfer the args to kwargs
    inputs = {
        "process_label": f"PythonJob<{task['name']}>",
        "function_source_code": orm.Str(function_source_code),
        "function_name": orm.Str(function_name),
        "code": code,
        "function_kwargs": function_kwargs,
        "upload_files": new_upload_files,
        "function_outputs": orm.List(function_outputs),
        "metadata": metadata,
        **kwargs,
    }
    return inputs


def prepare_for_shell_task(task: dict, inputs: dict) -> dict:
    """Prepare the inputs for ShellJob"""
    from aiida_shell.launch import prepare_shell_job_inputs
    import inspect

    # Retrieve the signature of `prepare_shell_job_inputs` to determine expected input parameters.
    signature = inspect.signature(prepare_shell_job_inputs)
    kwargs = {key: inputs.pop(key, None) for key in signature.parameters.keys()}
    inputs.update(prepare_shell_job_inputs(**kwargs))

    inputs.setdefault("metadata", {})
    inputs["metadata"].update({"call_link_label": task["name"]})
    return inputs



"""Parser for an `PythonJob` job."""
from aiida.parsers.parser import Parser
from aiida_workgraph.orm import general_serializer
from aiida.engine import ExitCode


class PythonParser(Parser):
    """Parser for an `PythonJob` job."""

    def parse(self, **kwargs):
        """Parse the contents of the output files stored in the `retrieved` output node.

        The function_outputs could be a namespce, e.g.,
        function_outputs=[
            {"identifier": "workgraph.namespace", "name": "add_multiply"},
            {"name": "add_multiply.add"},
            {"name": "add_multiply.multiply"},
            {"name": "minus"},
        ]
        """
        import pickle

        function_outputs = self.node.inputs.function_outputs.get_list()
        # function_outputs exclude ['_wait', '_outputs', 'remote_folder', 'remote_stash', 'retrieved']
        self.output_list = [
            data
            for data in function_outputs
            if data["name"]
            not in [
                "_wait",
                "_outputs",
                "remote_folder",
                "remote_stash",
                "retrieved",
                "exit_code",
            ]
        ]
        # first we remove nested outputs, e.g., "add_multiply.add"
        top_level_output_list = [
            output for output in self.output_list if "." not in output["name"]
        ]
        exit_code = 0
        try:
            with self.retrieved.base.repository.open("results.pickle", "rb") as handle:
                results = pickle.load(handle)
                if isinstance(results, tuple):
                    if len(top_level_output_list) != len(results):
                        self.exit_codes.ERROR_RESULT_OUTPUT_MISMATCH
                    for i in range(len(top_level_output_list)):
                        top_level_output_list[i]["value"] = self.serialize_output(
                            results[i], top_level_output_list[i]
                        )
                elif isinstance(results, dict) and len(top_level_output_list) > 1:
                    # pop the exit code if it exists
                    exit_code = results.pop("exit_code", 0)
                    for output in top_level_output_list:
                        if output.get("required", False):
                            if output["name"] not in results:
                                self.exit_codes.ERROR_MISSING_OUTPUT
                        output["value"] = self.serialize_output(
                            results.pop(output["name"]), output
                        )
                    # if there are any remaining results, raise an warning
                    if results:
                        self.logger.warning(
                            f"Found extra results that are not included in the output: {results.keys()}"
                        )
                elif isinstance(results, dict) and len(top_level_output_list) == 1:
                    exit_code = results.pop("exit_code", 0)
                    # if output name in results, use it
                    if top_level_output_list[0]["name"] in results:
                        top_level_output_list[0]["value"] = self.serialize_output(
                            results[top_level_output_list[0]["name"]],
                            top_level_output_list[0],
                        )
                    # otherwise, we assume the results is the output
                    else:
                        top_level_output_list[0]["value"] = self.serialize_output(
                            results, top_level_output_list[0]
                        )
                elif len(top_level_output_list) == 1:
                    # otherwise, we assume the results is the output
                    top_level_output_list[0]["value"] = self.serialize_output(
                        results, top_level_output_list[0]
                    )
                else:
                    raise ValueError(
                        "The number of results does not match the number of outputs."
                    )
                for output in top_level_output_list:
                    self.out(output["name"], output["value"])
                if exit_code:
                    if isinstance(exit_code, dict):
                        exit_code = ExitCode(exit_code["status"], exit_code["message"])
                    elif isinstance(exit_code, int):
                        exit_code = ExitCode(exit_code)
                    return exit_code
        except OSError:
            return self.exit_codes.ERROR_READING_OUTPUT_FILE
        except ValueError as exception:
            self.logger.error(exception)
            return self.exit_codes.ERROR_INVALID_OUTPUT

    def find_output(self, name):
        """Find the output with the given name."""
        for output in self.output_list:
            if output["name"] == name:
                return output
        return None

    def serialize_output(self, result, output):
        """Serialize outputs."""

        name = output["name"]
        if output.get("identifier", "Any").upper() == "WORKGRAPH.NAMESPACE":
            if isinstance(result, dict):
                serialized_result = {}
                for key, value in result.items():
                    full_name = f"{name}.{key}"
                    full_name_output = self.find_output(full_name)
                    if (
                        full_name_output
                        and full_name_output.get("identifier", "Any").upper()
                        == "WORKGRAPH.NAMESPACE"
                    ):
                        serialized_result[key] = self.serialize_output(
                            value, full_name_output
                        )
                    else:
                        serialized_result[key] = general_serializer(value)
                return serialized_result
            else:
                self.exit_codes.ERROR_INVALID_OUTPUT
        else:
            return general_serializer(result)



from aiida.orm import Data
from ase import Atoms
from ase.db.row import atoms2dict
import numpy as np

__all__ = ("AtomsData",)


class AtomsData(Data):
    """Data to represent a ASE Atoms."""

    _cached_atoms = None

    def __init__(self, value=None, **kwargs):
        """Initialise a `AtomsData` node instance.

        :param value: ASE Atoms instance to initialise the `AtomsData` node from
        """
        atoms = value or Atoms()
        super().__init__(**kwargs)
        data, keys = self.atoms2dict(atoms)
        self.base.attributes.set_many(data)
        self.base.attributes.set("keys", keys)

    @classmethod
    def atoms2dict(cls, atoms):
        data = atoms2dict(atoms)
        data.pop("unique_id")
        keys = list(data.keys())
        formula = atoms.get_chemical_formula()
        data = cls._convert_numpy_to_native(data)
        data["formula"] = formula
        data["symbols"] = atoms.get_chemical_symbols()
        return data, keys

    @classmethod
    def _convert_numpy_to_native(self, data):
        """Convert numpy types to Python native types for JSON compatibility."""
        for key, value in data.items():
            if isinstance(value, np.bool_):
                data[key] = bool(value)
            elif isinstance(value, np.ndarray):
                data[key] = value.tolist()
            elif isinstance(value, np.generic):
                data[key] = value.item()
        return data

    @property
    def value(self):
        keys = self.base.attributes.get("keys")
        data = self.base.attributes.get_many(keys)
        data = dict(zip(keys, data))
        return Atoms(**data)



"""The main `workgraph` click group."""
import click

from aiida_workgraph import __version__

from aiida.cmdline.groups import VerdiCommandGroup
from aiida.cmdline.params import options, types


# Pass the version explicitly to ``version_option`` otherwise editable installs can show the wrong version number
@click.group(
    cls=VerdiCommandGroup, context_settings={"help_option_names": ["--help", "-h"]}
)
@options.PROFILE(type=types.ProfileParamType(load_profile=True), expose_value=False)
@options.VERBOSITY()
@click.version_option(
    __version__,
    package_name="aiida_core",
    message="AiiDA-WorkGraph version %(version)s",
)
def workgraph():
    """The command line interface of AiiDA-WorkGraph."""



from typing import Any, Callable, Dict, List, Optional, Union, Tuple
from aiida_workgraph.utils import get_executor
from aiida.engine import calcfunction, workfunction, CalcJob, WorkChain
from aiida import orm
from aiida.orm.nodes.process.calculation.calcfunction import CalcFunctionNode
from aiida.orm.nodes.process.workflow.workfunction import WorkFunctionNode
from aiida.engine.processes.ports import PortNamespace
from aiida_workgraph.task import Task
from aiida_workgraph.utils import build_callable
import inspect

task_types = {
    CalcFunctionNode: "CALCFUNCTION",
    WorkFunctionNode: "WORKFUNCTION",
    CalcJob: "CALCJOB",
    WorkChain: "WORKCHAIN",
}

type_mapping = {
    "default": "workgraph.any",
    "namespace": "workgraph.namespace",
    int: "workgraph.int",
    float: "workgraph.float",
    str: "workgraph.string",
    bool: "workgraph.bool",
    orm.Int: "workgraph.aiida_int",
    orm.Float: "workgraph.aiida_float",
    orm.Str: "workgraph.aiida_string",
    orm.Bool: "workgraph.aiida_bool",
    orm.StructureData: "workgraph.aiida_structuredata",
}


def create_task(tdata):
    """Wrap create_node from node_graph to create a Task."""
    from node_graph.decorator import create_node

    tdata["type_mapping"] = type_mapping
    tdata["metadata"]["node_type"] = tdata["metadata"].pop("task_type")
    return create_node(tdata)


def add_input_recursive(
    inputs: List[List[Union[str, Dict[str, Any]]]],
    port: PortNamespace,
    args: List,
    kwargs: List,
    prefix: Optional[str] = None,
    required: bool = True,
) -> List[List[Union[str, Dict[str, Any]]]]:
    """Add input recursively."""
    if prefix is None:
        port_name = port.name
    else:
        port_name = f"{prefix}.{port.name}"
    required = port.required and required
    input_names = [input["name"] for input in inputs]
    if isinstance(port, PortNamespace):
        # TODO the default value is {} could cause problem, because the address of the dict is the same,
        # so if you change the value of one port, the value of all the ports of other tasks will be changed
        # consider to use None as default value
        if port_name not in input_names:
            inputs.append(
                {
                    "identifier": "workgraph.namespace",
                    "name": port_name,
                    "property": {"identifier": "workgraph.any", "default": None},
                }
            )
        if required:
            args.append(port_name)
        else:
            kwargs.append(port_name)
        for value in port.values():
            add_input_recursive(
                inputs, value, args, kwargs, prefix=port_name, required=required
            )
    else:
        if port_name not in input_names:
            # port.valid_type can be a single type or a tuple of types,
            # we only support single type for now
            if isinstance(port.valid_type, tuple) and len(port.valid_type) > 1:
                socket_type = "workgraph.any"
            if isinstance(port.valid_type, tuple) and len(port.valid_type) == 1:
                socket_type = type_mapping.get(port.valid_type[0], "workgraph.any")
            else:
                socket_type = type_mapping.get(port.valid_type, "workgraph.any")
            inputs.append({"identifier": socket_type, "name": port_name})
        if required:
            args.append(port_name)
        else:
            kwargs.append(port_name)
    return inputs


def add_output_recursive(
    outputs: List[List[Union[str, Dict[str, Any]]]],
    port: PortNamespace,
    prefix: Optional[str] = None,
    required: bool = True,
) -> List[List[Union[str, Dict[str, Any]]]]:
    """Add output recursively."""
    if prefix is None:
        port_name = port.name
    else:
        port_name = f"{prefix}.{port.name}"
    required = port.required and required
    output_names = [output["name"] for output in outputs]
    if isinstance(port, PortNamespace):
        # TODO the default value is {} could cause problem, because the address of the dict is the same,
        # so if you change the value of one port, the value of all the ports of other tasks will be changed
        # consider to use None as default value
        if port_name not in output_names:
            outputs.append({"identifier": "workgraph.namespace", "name": port_name})
        for value in port.values():
            add_output_recursive(outputs, value, prefix=port_name, required=required)
    else:
        if port_name not in output_names:
            outputs.append({"identifier": "workgraph.any", "name": port_name})
    return outputs


def build_task(
    executor: Union[Callable, str],
    inputs: Optional[List[str]] = None,
    outputs: Optional[List[str]] = None,
) -> Task:
    """Build task from executor."""
    from aiida_workgraph.workgraph import WorkGraph

    if isinstance(executor, WorkGraph):
        return build_task_from_workgraph(executor)
    elif isinstance(executor, str):
        (
            module,
            executor_name,
        ) = executor.rsplit(".", 1)
        executor, _ = get_executor({"module": module, "name": executor_name})
    if callable(executor):
        return build_task_from_callable(executor, inputs=inputs, outputs=outputs)


def build_task_from_callable(
    executor: Callable,
    inputs: Optional[List[str]] = None,
    outputs: Optional[List[str]] = None,
) -> Task:
    """Build task from a callable object.
    First, check if the executor is already a task.
    If not, check if it is a function or a class.
    If it is a function, build task from function.
    If it is a class, it only supports CalcJob and WorkChain.
    """
    from aiida_workgraph.task import Task

    # if it is already a task, return it
    if (
        hasattr(executor, "task")
        and inspect.isclass(executor.task)
        and issubclass(executor.task, Task)
        or inspect.isclass(executor)
        and issubclass(executor, Task)
    ):
        return executor
    tdata = {"metadata": {}}
    if inspect.isfunction(executor):
        # calcfunction and workfunction
        if getattr(executor, "node_class", False):
            tdata["metadata"]["task_type"] = task_types.get(
                executor.node_class, "NORMAL"
            )
            tdata["executor"] = executor
            return build_task_from_AiiDA(tdata, inputs=inputs, outputs=outputs)[0]
        else:
            tdata["metadata"]["task_type"] = "NORMAL"
            tdata["executor"] = executor
            return build_task_from_function(executor, inputs=inputs, outputs=outputs)
    else:
        if issubclass(executor, CalcJob):
            tdata["metadata"]["task_type"] = "CALCJOB"
            tdata["executor"] = executor
            return build_task_from_AiiDA(tdata, inputs=inputs, outputs=outputs)[0]
        elif issubclass(executor, WorkChain):
            tdata["metadata"]["task_type"] = "WORKCHAIN"
            tdata["executor"] = executor
            return build_task_from_AiiDA(tdata, inputs=inputs, outputs=outputs)[0]
    raise ValueError("The executor is not supported.")


def build_task_from_function(
    executor: Callable,
    inputs: Optional[List[str]] = None,
    outputs: Optional[List[str]] = None,
) -> Task:
    """Build task from function."""
    return TaskDecoratorCollection.decorator_task(inputs=inputs, outputs=outputs)(
        executor
    ).task


def build_task_from_AiiDA(
    tdata: Dict[str, Any],
    inputs: Optional[List[str]] = None,
    outputs: Optional[List[str]] = None,
) -> Task:
    """Register a task from a AiiDA component.
    For example: CalcJob, WorkChain, CalcFunction, WorkFunction."""

    tdata.setdefault("metadata", {})
    inputs = [] if inputs is None else inputs
    outputs = [] if outputs is None else outputs
    executor = tdata["executor"]
    spec = executor.spec()
    args = []
    kwargs = []
    user_defined_input_names = [input["name"] for input in inputs]
    for _key, port in spec.inputs.ports.items():
        add_input_recursive(inputs, port, args, kwargs, required=port.required)
    for _key, port in spec.outputs.ports.items():
        add_output_recursive(outputs, port, required=port.required)
    # Only check this for calcfunction and workfunction
    if inspect.isfunction(executor) and spec.inputs.dynamic:
        if hasattr(executor.process_class, "_varargs"):
            name = executor.process_class._varargs
        else:
            name = (
                executor.process_class._var_keyword
                or executor.process_class._var_positional
            )
        tdata["var_kwargs"] = name
        # if user already defined the var_args in the inputs, skip it
        if name not in [input["name"] for input in inputs]:
            inputs.append(
                {
                    "identifier": "workgraph.any",
                    "name": name,
                    "property": {"identifier": "workgraph.any", "default": {}},
                }
            )
        # When the input is dyanmic, if user defines some input names does not included in the args and kwargs,
        # which means the user define the input names manually, we must add them to the kwargs
        for key in user_defined_input_names:
            if key not in args and key not in kwargs:
                if key == name:
                    continue
                kwargs.append(key)

    # TODO In order to reload the WorkGraph from process, "is_pickle" should be True
    # so I pickled the function here, but this is not necessary
    # we need to update the node_graph to support the path and name of the function
    tdata["identifier"] = tdata.pop("identifier", tdata["executor"].__name__)
    tdata["executor"] = build_callable(executor)
    tdata["executor"]["type"] = tdata["metadata"]["task_type"]
    if tdata["metadata"]["task_type"].upper() in ["CALCFUNCTION", "WORKFUNCTION"]:
        outputs = (
            [{"identifier": "workgraph.any", "name": "result"}]
            if not outputs
            else outputs
        )
    # add built-in sockets
    outputs.append({"identifier": "workgraph.any", "name": "_outputs"})
    outputs.append({"identifier": "workgraph.any", "name": "_wait"})
    inputs.append({"identifier": "workgraph.any", "name": "_wait", "link_limit": 1e6})
    tdata["metadata"]["node_class"] = {"module": "aiida_workgraph.task", "name": "Task"}
    tdata["args"] = args
    tdata["kwargs"] = kwargs
    tdata["inputs"] = inputs
    tdata["outputs"] = outputs
    task = create_task(tdata)
    task.is_aiida_component = True
    return task, tdata


def build_pythonjob_task(func: Callable) -> Task:
    """Build PythonJob task from function."""
    from aiida_workgraph.calculations.python import PythonJob
    from copy import deepcopy

    # if the function is not a task, build a task from the function
    if not hasattr(func, "node"):
        TaskDecoratorCollection.decorator_task()(func)
    if func.node.node_type.upper() == "GRAPH_BUILDER":
        raise ValueError(
            "GraphBuilder task cannot be run remotely. Please remove 'PythonJob'."
        )
    tdata = {
        "metadata": {"task_type": "PYTHONJOB"},
        "executor": PythonJob,
    }
    _, tdata_py = build_task_from_AiiDA(tdata)
    tdata = deepcopy(func.tdata)
    function_kwargs = [
        name for name in tdata["inputs"] if name not in ["_wait", "_outputs"]
    ]
    # merge the inputs and outputs from the PythonJob task to the function task
    # skip the already existed inputs and outputs
    for input in [
        {"identifier": "workgraph.string", "name": "computer"},
        {"identifier": "workgraph.string", "name": "code_label"},
        {"identifier": "workgraph.string", "name": "code_path"},
        {"identifier": "workgraph.string", "name": "prepend_text"},
    ]:
        input["list_index"] = len(tdata["inputs"]) + 1
        tdata["inputs"][input["name"]] = input
    for name, input in tdata_py["inputs"].items():
        if name not in tdata["inputs"]:
            input["list_index"] = len(tdata["inputs"]) + 1
            tdata["inputs"][name] = input
    for name, output in tdata_py["outputs"].items():
        if name not in tdata["outputs"]:
            output["list_index"] = len(tdata["outputs"]) + 1
            tdata["outputs"][name] = output
    for output in [{"identifier": "workgraph.any", "name": "exit_code"}]:
        output["list_index"] = len(tdata["outputs"]) + 1
        tdata["outputs"][output["name"]] = output
    # change "copy_files" link_limit to 1e6
    tdata["inputs"]["copy_files"]["link_limit"] = 1e6
    # append the kwargs of the PythonJob task to the function task
    kwargs = tdata["kwargs"]
    kwargs.extend(["computer", "code_label", "code_path", "prepend_text"])
    kwargs.extend(tdata_py["kwargs"])
    tdata["kwargs"] = kwargs
    tdata["metadata"]["task_type"] = "PYTHONJOB"
    tdata["identifier"] = "workgraph.pythonjob"
    tdata["metadata"]["node_class"] = {
        "module": "aiida_workgraph.tasks.pythonjob",
        "name": "PythonJob",
    }
    task = create_task(tdata)
    task.is_aiida_component = True
    task.function_kwargs = function_kwargs
    return task, tdata


def build_shelljob_task(
    nodes: dict = None, outputs: list = None, parser_outputs: list = None
) -> Task:
    """Build ShellJob with custom inputs and outputs."""
    from aiida_shell.calculations.shell import ShellJob
    from aiida_shell.parsers.shell import ShellParser
    from node_graph.socket import NodeSocket

    tdata = {
        "metadata": {"task_type": "SHELLJOB"},
        "executor": ShellJob,
    }
    _, tdata = build_task_from_AiiDA(tdata)
    # create input sockets for the nodes, if it is linked other sockets
    links = {}
    inputs = []
    nodes = {} if nodes is None else nodes
    keys = list(nodes.keys())
    for key in keys:
        inputs.append({"identifier": "workgraph.any", "name": f"nodes.{key}"})
        # input is a output of another task, we make a link
        if isinstance(nodes[key], NodeSocket):
            links[f"nodes.{key}"] = nodes[key]
            # Output socket itself is not a value, so we remove the key from the nodes
            nodes.pop(key)
    for input in inputs:
        if input["name"] not in tdata["inputs"]:
            input["list_index"] = len(tdata["inputs"]) + 1
            tdata["inputs"][input["name"]] = input
            tdata["kwargs"].append(input["name"])
    # Extend the outputs
    for output in [
        {"identifier": "workgraph.any", "name": "stdout"},
        {"identifier": "workgraph.any", "name": "stderr"},
    ]:
        output["list_index"] = len(tdata["outputs"]) + 1
        tdata["outputs"][output["name"]] = output
    outputs = [] if outputs is None else outputs
    parser_outputs = [] if parser_outputs is None else parser_outputs
    outputs = [
        {"identifier": "workgraph.any", "name": ShellParser.format_link_label(output)}
        for output in outputs
    ]
    outputs.extend(parser_outputs)
    # add user defined outputs
    for output in outputs:
        if output["name"] not in tdata["outputs"]:
            output["list_index"] = len(tdata["outputs"]) + 1
            tdata["outputs"][output["name"]] = output
    #
    tdata["identifier"] = "ShellJob"
    for input in [
        {"identifier": "workgraph.any", "name": "command"},
        {"identifier": "workgraph.any", "name": "resolve_command"},
    ]:
        input["list_index"] = len(tdata["inputs"]) + 1
        tdata["inputs"][input["name"]] = input
    tdata["kwargs"].extend(["command", "resolve_command"])
    tdata["metadata"]["task_type"] = "SHELLJOB"
    task = create_task(tdata)
    task.is_aiida_component = True
    return task, tdata, links


def build_task_from_workgraph(wg: any) -> Task:
    """Build task from workgraph."""
    from aiida.orm.utils.serialize import serialize

    tdata = {"metadata": {"task_type": "workgraph"}}
    inputs = []
    outputs = []
    group_outputs = []
    # add all the inputs/outputs from the tasks in the workgraph
    for task in wg.tasks:
        # inputs
        inputs.append(
            {
                "identifier": "workgraph.namespace",
                "name": f"{task.name}",
            }
        )
        for socket in task.inputs:
            if socket.name == "_wait":
                continue
            inputs.append(
                {"identifier": socket.identifier, "name": f"{task.name}.{socket.name}"}
            )
        # outputs
        outputs.append(
            {
                "identifier": "workgraph.namespace",
                "name": f"{task.name}",
            }
        )
        for socket in task.outputs:
            if socket.name in ["_wait", "_outputs"]:
                continue
            outputs.append(
                {"identifier": socket.identifier, "name": f"{task.name}.{socket.name}"}
            )
            group_outputs.append(
                {
                    "name": f"{task.name}.{socket.name}",
                    "from": f"{task.name}.{socket.name}",
                }
            )
    kwargs = [input["name"] for input in inputs]
    # add built-in sockets
    outputs.append({"identifier": "workgraph.any", "name": "_outputs"})
    outputs.append({"identifier": "workgraph.any", "name": "_wait"})
    inputs.append({"identifier": "workgraph.any", "name": "_wait", "link_limit": 1e6})
    tdata["metadata"]["node_class"] = {"module": "aiida_workgraph.task", "name": "Task"}
    tdata["kwargs"] = kwargs
    tdata["inputs"] = inputs
    tdata["outputs"] = outputs
    tdata["identifier"] = wg.name
    executor = {
        "module": "aiida_workgraph.engine.workgraph",
        "name": "WorkGraphEngine",
        "wgdata": serialize(wg.to_dict(store_nodes=True)),
        "type": tdata["metadata"]["task_type"],
        "is_pickle": False,
    }
    tdata["metadata"]["group_outputs"] = group_outputs
    tdata["executor"] = executor
    task = create_task(tdata)
    return task


def nonfunctional_usage(callable: Callable):
    """
    This is a decorator for a decorator factory (a function that returns a decorator).
    It allows the usage of the decorator factory in a nonfunctional way. So a decorator
    factory that has been decorated by this decorator that could only be used befor like
    this

    .. code-block:: python

        @decorator_factory()
        def foo():
            pass

    can now be also used like this

    .. code-block:: python

        @decorator_factory
        def foo():
            pass

    """

    def decorator_task_wrapper(*args, **kwargs):
        if len(args) == 1 and isinstance(args[0], Callable) and len(kwargs) == 0:
            return callable()(args[0])
        else:
            return callable(*args, **kwargs)

    return decorator_task_wrapper


def generate_tdata(
    func: Callable,
    identifier: str,
    inputs: List[Tuple[str, str]],
    outputs: List[Tuple[str, str]],
    properties: List[Tuple[str, str]],
    catalog: str,
    task_type: str,
    group_inputs: List[Tuple[str, str]] = None,
    group_outputs: List[Tuple[str, str]] = None,
    additional_data: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    """Generate task data for creating a task."""
    from node_graph.decorator import generate_input_sockets

    args, kwargs, var_args, var_kwargs, _inputs = generate_input_sockets(
        func, inputs, properties, type_mapping=type_mapping
    )
    task_outputs = outputs
    # add built-in sockets
    _inputs.append({"identifier": "workgraph.any", "name": "_wait", "link_limit": 1e6})
    task_outputs.append({"identifier": "workgraph.any", "name": "_wait"})
    task_outputs.append({"identifier": "workgraph.any", "name": "_outputs"})
    tdata = {
        "identifier": identifier,
        "args": args,
        "kwargs": kwargs,
        "var_args": var_args,
        "var_kwargs": var_kwargs,
        "metadata": {
            "task_type": task_type,
            "catalog": catalog,
            "node_class": {"module": "aiida_workgraph.task", "name": "Task"},
            "group_inputs": group_inputs or [],
            "group_outputs": group_outputs or [],
        },
        "properties": properties,
        "inputs": _inputs,
        "outputs": task_outputs,
    }
    tdata["executor"] = build_callable(func)
    if additional_data:
        tdata.update(additional_data)
    return tdata


class TaskDecoratorCollection:
    """Collection of task decorators."""

    # decorator with arguments indentifier, args, kwargs, properties, inputs, outputs, executor
    @staticmethod
    @nonfunctional_usage
    def decorator_task(
        identifier: Optional[str] = None,
        task_type: str = "Normal",
        properties: Optional[List[Tuple[str, str]]] = None,
        inputs: Optional[List[Tuple[str, str]]] = None,
        outputs: Optional[List[Tuple[str, str]]] = None,
        error_handlers: Optional[List[Dict[str, Any]]] = None,
        catalog: str = "Others",
    ) -> Callable:
        """Generate a decorator that register a function as a task.

        Attributes:
            indentifier (str): task identifier
            catalog (str): task catalog
            args (list): task args
            kwargs (dict): task kwargs
            properties (list): task properties
            inputs (list): task inputs
            outputs (list): task outputs
        """

        def decorator(func):
            nonlocal identifier, task_type

            if identifier is None:
                identifier = func.__name__

            # Determine task_type based on AiiDA's node classes
            task_type = task_type
            if hasattr(func, "node_class"):
                task_type = task_types.get(func.node_class, task_type)
            tdata = generate_tdata(
                func,
                identifier,
                inputs or [],
                outputs or [{"identifier": "workgraph.any", "name": "result"}],
                properties or [],
                catalog,
                task_type,
            )
            task = create_task(tdata)
            task._error_handlers = error_handlers
            func.identifier = identifier
            func.task = func.node = task
            func.tdata = tdata
            return func

        return decorator

    # decorator with arguments indentifier, args, kwargs, properties, inputs, outputs, executor
    @staticmethod
    @nonfunctional_usage
    def decorator_graph_builder(
        identifier: Optional[str] = None,
        properties: Optional[List[Tuple[str, str]]] = None,
        inputs: Optional[List[Tuple[str, str]]] = None,
        outputs: Optional[List[Tuple[str, str]]] = None,
        catalog: str = "Others",
    ) -> Callable:
        """Generate a decorator that register a function as a graph_builder task.
        Attributes:
            indentifier (str): task identifier
            catalog (str): task catalog
            properties (list): task properties
            inputs (list): task inputs
            outputs (list): task outputs
        """

        outputs = outputs or []

        def decorator(func):
            nonlocal identifier, inputs, outputs

            if identifier is None:
                identifier = func.__name__
            # use cloudpickle to serialize function
            func.identifier = identifier

            task_outputs = [
                {"identifier": "workgraph.any", "name": output["name"]}
                for output in outputs
            ]
            # print(task_inputs, task_outputs)
            #
            task_type = "graph_builder"
            tdata = generate_tdata(
                func,
                identifier,
                inputs or [],
                task_outputs,
                properties or [],
                catalog,
                task_type,
                group_inputs=inputs,
                group_outputs=outputs,
            )

            task = create_task(tdata)
            func.task = func.node = task
            return func

        return decorator

    @staticmethod
    @nonfunctional_usage
    def calcfunction(**kwargs: Any) -> Callable:
        def decorator(func):
            # First, apply the calcfunction decorator
            func_decorated = calcfunction(func)
            # Then, apply task decorator
            task_decorated = build_task_from_callable(
                func_decorated,
                inputs=kwargs.get("inputs", []),
                outputs=kwargs.get("outputs", []),
            )
            identifier = kwargs.get("identifier", None)
            func_decorated.identifier = identifier if identifier else func.__name__
            func_decorated.task = func_decorated.node = task_decorated
            return func_decorated

        return decorator

    @staticmethod
    @nonfunctional_usage
    def workfunction(**kwargs: Any) -> Callable:
        def decorator(func):
            # First, apply the workfunction decorator
            func_decorated = workfunction(func)
            task_decorated = build_task_from_callable(
                func_decorated,
                inputs=kwargs.get("inputs", []),
                outputs=kwargs.get("outputs", []),
            )
            identifier = kwargs.get("identifier", None)
            func_decorated.identifier = identifier if identifier else func.__name__
            func_decorated.task = func_decorated.node = task_decorated

            return func_decorated

        return decorator

    @staticmethod
    @nonfunctional_usage
    def pythonjob(**kwargs: Any) -> Callable:
        def decorator(func):
            # first create a task from the function
            task_decorated = build_task_from_callable(
                func,
                inputs=kwargs.get("inputs", []),
                outputs=kwargs.get("outputs", []),
            )
            # then build a PythonJob task from the function task
            task_decorated, _ = build_pythonjob_task(func)
            func.identifier = "PythonJob"
            func.task = func.node = task_decorated
            task_decorated._error_handlers = kwargs.get("error_handlers", [])

            return func

        return decorator

    @staticmethod
    @nonfunctional_usage
    def awaitable(**kwargs: Any) -> Callable:
        def decorator(func):
            # Then, apply task decorator
            task_decorated = build_task_from_callable(
                func,
                inputs=kwargs.get("inputs", []),
                outputs=kwargs.get("outputs", []),
            )
            task_decorated.node_type = "awaitable"
            func.identifier = "awaitable"
            func.task = func.node = task_decorated
            return func

        return decorator

    @staticmethod
    @nonfunctional_usage
    def monitor(**kwargs: Any) -> Callable:
        def decorator(func):
            # Then, apply task decorator
            task_decorated = build_task_from_callable(
                func,
                inputs=kwargs.get(
                    "inputs", [{"identifier": "workgraph.any", "name": "interval"}]
                ),
                outputs=kwargs.get("outputs", []),
            )
            task_decorated.node_type = "monitor"
            func.identifier = "monitor"
            func.task = func.node = task_decorated
            return func

        return decorator

    # Making decorator_task accessible as 'task'
    task = decorator_task

    # Making decorator_graph_builder accessible as 'graph_builder'
    graph_builder = decorator_graph_builder

    def __call__(self, *args, **kwargs):
        # This allows using '@task' to directly apply the decorator_task functionality
        if len(args) == 1 and isinstance(args[0], Callable) and len(kwargs) == 0:
            return self.decorator_task()(args[0])
        else:
            return self.decorator_task(*args, **kwargs)


task = TaskDecoratorCollection()



from aiida_workgraph.task import Task


class TimeMonitor(Task):
    """Monitor the time"""

    identifier = "workgraph.time_monitor"
    name = "TimeMonitor"
    node_type = "MONITOR"
    catalog = "Monitor"
    args = ["datetime"]

    _executor = {
        "module": "aiida_workgraph.executors.monitors",
        "name": "time_monitor",
    }
    kwargs = ["interval", "timeout"]

    def create_sockets(self) -> None:
        self.inputs.clear()
        self.outputs.clear()
        self.inputs.new("workgraph.any", "datetime")
        inp = self.inputs.new("workgraph.any", "interval")
        inp.add_property("workgraph.any", default=1.0)
        inp = self.inputs.new("workgraph.any", "timeout")
        inp.add_property("workgraph.any", default=86400.0)
        inp = self.inputs.new("workgraph.any", "_wait")
        inp.link_limit = 100000
        self.outputs.new("workgraph.any", "result")
        self.outputs.new("workgraph.any", "_wait")


class FileMonitor(Task):
    """Monitor the file"""

    identifier = "workgraph.file_monitor"
    name = "FileMonitor"
    node_type = "MONITOR"
    catalog = "Monitor"
    args = ["filepath"]

    _executor = {
        "module": "aiida_workgraph.executors.monitors",
        "name": "file_monitor",
    }
    kwargs = ["interval", "timeout"]

    def create_sockets(self) -> None:
        self.inputs.clear()
        self.outputs.clear()
        self.inputs.new("workgraph.any", "filepath")
        inp = self.inputs.new("workgraph.any", "interval")
        inp.add_property("workgraph.any", default=1.0)
        inp = self.inputs.new("workgraph.any", "timeout")
        inp.add_property("workgraph.any", default=86400.0)
        inp = self.inputs.new("workgraph.any", "_wait")
        inp.link_limit = 100000
        self.outputs.new("workgraph.any", "result")
        self.outputs.new("workgraph.any", "_wait")


class TaskMonitor(Task):
    """Monitor the file"""

    identifier = "workgraph.task_monitor"
    name = "TaskMonitor"
    node_type = "MONITOR"
    catalog = "Monitor"
    args = ["task_name"]

    _executor = {
        "module": "aiida_workgraph.executors.monitors",
        "name": "task_monitor",
    }
    kwargs = ["interval", "timeout", "workgraph_pk", "workgraph_name"]

    def create_sockets(self) -> None:
        self.inputs.clear()
        self.outputs.clear()
        self.inputs.new("workgraph.any", "workgraph_pk")
        self.inputs.new("workgraph.any", "workgraph_name")
        self.inputs.new("workgraph.any", "task_name")
        inp = self.inputs.new("workgraph.any", "interval")
        inp.add_property("workgraph.any", default=1.0)
        inp = self.inputs.new("workgraph.any", "timeout")
        inp.add_property("workgraph.any", default=86400.0)
        inp = self.inputs.new("workgraph.any", "_wait")
        inp.link_limit = 100000
        self.outputs.new("workgraph.any", "result")
        self.outputs.new("workgraph.any", "_wait")



"""Sub commands of the ``verdi`` command line interface.

The commands need to be imported here for them to be registered with the top-level command group.
"""
from aiida_workgraph.cli import cmd_graph
from aiida_workgraph.cli import cmd_web
from aiida_workgraph.cli import cmd_task


__all__ = ["cmd_graph", "cmd_web", "cmd_task"]



from typing import Any, Dict
from aiida_workgraph.task import Task, TaskCollection


class Zone(Task):
    """
    Extend the Task class to include a 'children' attribute.
    """

    identifier = "workgraph.zone"
    name = "Zone"
    node_type = "ZONE"
    catalog = "Control"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.children = TaskCollection(parent=self)

    def create_sockets(self) -> None:
        self.inputs.clear()
        self.outputs.clear()
        inp = self.inputs.new("workgraph.any", "_wait")
        inp.link_limit = 100000
        self.outputs.new("workgraph.any", "_wait")

    def to_dict(self) -> Dict[str, Any]:
        tdata = super().to_dict()
        tdata["children"] = [task.name for task in self.children]
        return tdata

    def from_dict(self, data: Dict[str, Any]) -> None:
        super().from_dict(data)
        self.children.add(data.get("children", []))


class While(Zone):
    """While"""

    identifier = "workgraph.while"
    name = "While"
    node_type = "WHILE"
    catalog = "Control"
    kwargs = ["max_iterations", "conditions"]

    def create_sockets(self) -> None:
        self.inputs.clear()
        self.outputs.clear()
        inp = self.inputs.new("workgraph.any", "_wait")
        inp.link_limit = 100000
        inp = self.inputs.new("node_graph.int", "max_iterations")
        inp.add_property("node_graph.int", default=10000)
        inp = self.inputs.new("workgraph.any", "conditions")
        inp.link_limit = 100000
        self.outputs.new("workgraph.any", "_wait")


class If(Zone):
    """If task"""

    identifier = "workgraph.if"
    name = "If"
    node_type = "IF"
    catalog = "Control"
    kwargs = ["conditions", "invert_condition"]

    def create_sockets(self) -> None:
        self.inputs.clear()
        self.outputs.clear()
        inp = self.inputs.new("workgraph.any", "_wait")
        inp.link_limit = 100000
        self.inputs.new("workgraph.any", "conditions")
        self.inputs.new("workgraph.any", "invert_condition")
        self.outputs.new("workgraph.any", "_wait")


class Gather(Task):
    """Gather"""

    identifier = "workgraph.aiida_gather"
    name = "Gather"
    node_type = "WORKCHAIN"
    catalog = "Control"

    _executor = {
        "module": "aiida_workgraph.executors.builtins",
        "name": "GatherWorkChain",
    }
    kwargs = ["datas"]

    def create_sockets(self) -> None:
        self.inputs.clear()
        self.outputs.clear()
        inp = self.inputs.new("workgraph.any", "datas")
        inp.link_limit = 100000
        self.outputs.new("workgraph.any", "result")


class ToContext(Task):
    """ToContext"""

    identifier = "workgraph.to_context"
    name = "ToContext"
    node_type = "TO_CONTEXT"
    catalog = "Control"
    args = ["key", "value"]

    def create_sockets(self) -> None:
        self.inputs.clear()
        self.outputs.clear()
        self.inputs.new("workgraph.any", "key")
        inp = self.inputs.new("workgraph.any", "_wait")
        inp.link_limit = 100000
        self.inputs.new("workgraph.any", "value")
        self.outputs.new("workgraph.any", "_wait")


class FromContext(Task):
    """FromContext"""

    identifier = "workgraph.from_context"
    name = "FromContext"
    node_type = "FROM_CONTEXT"
    catalog = "Control"
    args = ["key"]

    def create_sockets(self) -> None:
        self.inputs.clear()
        self.outputs.clear()
        self.inputs.new("workgraph.any", "key")
        inp = self.inputs.new("workgraph.any", "_wait")
        inp.link_limit = 100000
        self.outputs.new("workgraph.any", "result")
        self.outputs.new("workgraph.any", "_wait")


class AiiDAInt(Task):
    identifier = "workgraph.aiida_int"
    name = "AiiDAInt"
    node_type = "data"
    catalog = "Test"

    _executor = {
        "module": "aiida.orm",
        "name": "Int",
    }
    args = ["value"]

    def create_sockets(self) -> None:
        inp = self.inputs.new("workgraph.any", "value", default=0.0)
        inp.add_property("workgraph.aiida_int", default=1.0)
        self.outputs.new("workgraph.aiida_int", "result")


class AiiDAFloat(Task):
    identifier = "workgraph.aiida_float"
    name = "AiiDAFloat"
    node_type = "data"
    catalog = "Test"

    _executor = {
        "module": "aiida.orm",
        "name": "Float",
    }
    args = ["value"]

    def create_sockets(self) -> None:
        self.inputs.new("workgraph.aiida_float", "value", default=0.0)
        self.outputs.new("workgraph.aiida_float", "result")


class AiiDAString(Task):
    identifier = "workgraph.aiida_string"
    name = "AiiDAString"
    node_type = "data"
    catalog = "Test"

    _executor = {
        "module": "aiida.orm",
        "name": "Str",
    }
    args = ["value"]

    def create_sockets(self) -> None:
        self.inputs.new("AiiDAString", "value", default="")
        self.outputs.new("AiiDAString", "result")


class AiiDAList(Task):
    identifier = "workgraph.aiida_list"
    name = "AiiDAList"
    node_type = "data"
    catalog = "Test"

    _executor = {
        "module": "aiida.orm",
        "name": "List",
    }
    args = ["value"]

    def create_properties(self) -> None:
        self.properties.new("BaseList", "value", default=[])

    def create_sockets(self) -> None:
        self.outputs.new("workgraph.any", "Parameters")


class AiiDADict(Task):
    identifier = "workgraph.aiida_dict"
    name = "AiiDADict"
    node_type = "data"
    catalog = "Test"

    _executor = {
        "module": "aiida.orm",
        "name": "Dict",
    }
    args = ["value"]

    def create_properties(self) -> None:
        self.properties.new("BaseDict", "value", default={})

    def create_sockets(self) -> None:
        self.outputs.new("workgraph.any", "Parameters")


class AiiDANode(Task):
    """AiiDANode"""

    identifier = "workgraph.aiida_node"
    name = "AiiDANode"
    node_type = "node"
    catalog = "Test"

    _executor = {
        "module": "aiida.orm",
        "name": "load_node",
    }
    kwargs = ["identifier", "pk", "uuid", "label"]

    def create_properties(self) -> None:
        pass

    def create_sockets(self) -> None:
        self.inputs.clear()
        self.outputs.clear()
        self.inputs.new("workgraph.any", "identifier")
        self.inputs.new("workgraph.any", "pk")
        self.inputs.new("workgraph.any", "uuid")
        self.inputs.new("workgraph.any", "label")
        self.outputs.new("workgraph.any", "node")


class AiiDACode(Task):
    """AiiDACode"""

    identifier = "workgraph.aiida_code"
    name = "AiiDACode"
    node_type = "node"
    catalog = "Test"

    _executor = {
        "module": "aiida.orm",
        "name": "load_code",
    }
    kwargs = ["identifier", "pk", "uuid", "label"]

    def create_sockets(self) -> None:
        self.inputs.clear()
        self.outputs.clear()
        self.inputs.new("workgraph.any", "identifier")
        self.inputs.new("workgraph.any", "pk")
        self.inputs.new("workgraph.any", "uuid")
        self.inputs.new("workgraph.any", "label")
        self.outputs.new("workgraph.any", "Code")


class Select(Task):
    """Select"""

    identifier = "workgraph.select"
    name = "Select"
    node_type = "Normal"
    catalog = "Control"

    _executor = {
        "module": "aiida_workgraph.executors.builtins",
        "name": "select",
    }
    args = ["condition", "true", "false"]

    def create_sockets(self) -> None:
        self.inputs.clear()
        self.outputs.clear()
        self.inputs.new("workgraph.any", "condition")
        self.inputs.new("workgraph.any", "true")
        self.inputs.new("workgraph.any", "false")
        inp = self.inputs.new("workgraph.any", "_wait")
        inp.link_limit = 100000
        self.outputs.new("workgraph.any", "result")
        self.outputs.new("workgraph.any", "_wait")



from typing import Any, Dict, Optional, Union, Callable, List
from aiida.engine.processes import Process
from aiida import orm
from aiida.common.exceptions import NotExistent
from aiida.engine.runners import Runner


def build_callable(obj: Callable) -> Dict[str, Any]:
    """
    Build the executor data from the callable. This will either serialize the callable
    using cloudpickle if it's a local or lambda function, or store its module and name
    if it's a globally defined callable (function or class).

    Args:
        obj (Callable): The callable to be serialized or referenced.

    Returns:
        Dict[str, Any]: A dictionary containing the serialized callable or a reference
                        to its module and name.
    """
    import types
    from aiida_workgraph.orm.function_data import PickledFunction

    # Check if the callable is a function or class
    if isinstance(obj, (types.FunctionType, type)):
        # Check if callable is nested (contains dots in __qualname__ after the first segment)
        if obj.__module__ == "__main__" or "." in obj.__qualname__.split(".", 1)[-1]:
            # Local or nested callable, so pickle the callable
            executor = PickledFunction.build_callable(obj)
        else:
            # Global callable (function/class), store its module and name for reference
            executor = {
                "module": obj.__module__,
                "name": obj.__name__,
                "is_pickle": False,
            }
    elif isinstance(obj, PickledFunction) or isinstance(obj, dict):
        executor = obj
    else:
        raise TypeError("Provided object is not a callable function or class.")
    return executor


def get_sorted_names(data: dict) -> list:
    """Get the sorted names from a dictionary."""
    print("data: ", data)
    sorted_names = [
        name
        for name, _ in sorted(
            ((name, item["list_index"]) for name, item in data.items()),
            key=lambda x: x[1],
        )
    ]
    return sorted_names


def store_nodes_recursely(data: Any) -> None:
    """Recurse through a data structure and store any unstored nodes that are found along the way
    :param data: a data structure potentially containing unstored nodes
    """
    from aiida.orm import Node
    import collections.abc

    if isinstance(data, Node) and not data.is_stored:
        data.store()
    elif isinstance(data, collections.abc.Mapping):
        for _, value in data.items():
            store_nodes_recursely(value)
    elif isinstance(data, collections.abc.Sequence) and not isinstance(data, str):
        for value in data:
            store_nodes_recursely(value)


def get_executor(data: Dict[str, Any]) -> Union[Process, Any]:
    """Import executor from path and return the executor and type."""
    import importlib
    from aiida.plugins import CalculationFactory, WorkflowFactory, DataFactory

    data = data or {}
    is_pickle = data.get("is_pickle", False)
    type = data.get("type", "function")
    if is_pickle:
        import cloudpickle as pickle

        try:
            executor = pickle.loads(data["executor"])
        except Exception as e:
            print("Error in loading executor: ", e)
            executor = None
    else:
        if type == "WorkflowFactory":
            executor = WorkflowFactory(data["name"])
        elif type == "CalculationFactory":
            executor = CalculationFactory(data["name"])
        elif type == "DataFactory":
            executor = DataFactory(data["name"])
        elif not data.get("name", None) and not data.get("module", None):
            executor = None
        else:
            module = importlib.import_module("{}".format(data.get("module", "")))
            executor = getattr(module, data["name"])

    return executor, type


def create_data_node(executor: orm.Data, args: list, kwargs: dict) -> orm.Node:
    """Create an AiiDA data node from the executor and args and kwargs."""
    from aiida import orm

    # print("Create data node: ", executor, args, kwargs)
    extras = kwargs.pop("extras", {})
    repository_metadata = kwargs.pop("repository_metadata", {})
    if issubclass(executor, (orm.BaseType, orm.Dict)):
        data_node = executor(*args)
    else:
        data_node = executor(*args)
        data_node.base.attributes.set_many(kwargs)
    data_node.base.extras.set_many(extras)
    data_node.base.repository.repository_metadata = repository_metadata
    data_node.store()
    return data_node


def get_nested_dict(d: Dict, name: str, **kwargs) -> Any:
    """Get the value from a nested dictionary.
    If default is provided, return the default value if the key is not found.
    Otherwise, raise ValueError.
    For example:
    d = {"base": {"pw": {"parameters": 2}}}
    name = "base.pw.parameters"
    """
    from aiida.orm.utils.managers import NodeLinksManager

    keys = name.split(".")
    current = d
    for key in keys:
        if key not in current:
            if "default" in kwargs:
                return kwargs.get("default")
            else:
                if isinstance(current, dict):
                    avaiable_keys = current.keys()
                elif isinstance(current, NodeLinksManager):
                    avaiable_keys = list(current._get_keys())
                else:
                    avaiable_keys = []
                raise ValueError(f"{name} not exist. Available keys: {avaiable_keys}")
        current = current[key]
    return current


def merge_dicts(existing: Any, new: Any) -> Any:
    """Recursively merges two dictionaries."""
    if isinstance(existing, dict) and isinstance(new, dict):
        for k, v in new.items():
            if k in existing and isinstance(existing[k], dict) and isinstance(v, dict):
                merge_dicts(existing[k], v)
            else:
                existing[k] = v
    else:
        return new


def update_nested_dict(d: Optional[Dict[str, Any]], key: str, value: Any) -> None:
    """
    Update or create a nested dictionary structure based on a dotted key path.

    This function allows updating a nested dictionary or creating one if `d` is `None`.
    Given a dictionary and a key path (e.g., "base.pw.parameters"), it will traverse
    or create the necessary nested structure to set the provided value at the specified
    key location. If intermediate dictionaries do not exist, they will be created.
    If the resulting dictionary is empty, it is set to `None`.

    Args:
        d (Dict[str, Any] | None): The dictionary to update, which can be `None`.
                                   If `None`, an empty dictionary will be created.
        key (str): A dotted key path string representing the nested structure.
        value (Any): The value to set at the specified key.

    Example:
        d = None
        key = "base.pw.parameters"
        value = 2
        After running:
            update_nested_dict(d, key, value)
        The result will be:
            d = {"base": {"pw": {"parameters": 2}}}

    Edge Case:
        If the resulting dictionary is empty after the update, it will be set to `None`.

    """

    keys = key.split(".")
    current = d if d is not None else {}
    for k in keys[:-1]:
        current = current.setdefault(k, {})
    # Handle merging instead of overwriting
    last_key = keys[-1]
    if (
        last_key in current
        and isinstance(current[last_key], dict)
        and isinstance(value, dict)
    ):
        merge_dicts(current[last_key], value)
    else:
        current[last_key] = value
    # if current is empty, set it to None
    if not current:
        current = None
    return current


def is_empty(value: Any) -> bool:
    """Check if the provided value is an empty collection."""
    import numpy as np

    if isinstance(value, np.ndarray):
        return value.size == 0
    elif isinstance(value, (dict, list, set, tuple)):
        return not value
    return False


def update_nested_dict_with_special_keys(data: Dict[str, Any]) -> Dict[str, Any]:
    """Remove None and empty value"""
    # data = {k: v for k, v in data.items() if v is not None and not is_empty(v)}
    data = {k: v for k, v in data.items() if v is not None}
    #
    special_keys = [k for k in data.keys() if "." in k]
    for key in special_keys:
        value = data.pop(key)
        update_nested_dict(data, key, value)
    return data


def merge_properties(wgdata: Dict[str, Any]) -> None:
    """Merge sub properties to the root properties.
        {
            "base.pw.parameters": 2,
            "base.pw.code": 1,
        }
        after merge:
        {"base": {"pw": {"parameters": 2,
                        "code": 1}}
    So that no "." in the key name.
    """
    for _, task in wgdata["tasks"].items():
        for key, prop in task["properties"].items():
            if "." in key and prop["value"] not in [None, {}]:
                root, key = key.split(".", 1)
                root_prop = task["properties"][root]
                update_nested_dict(root_prop["value"], key, prop["value"])
                prop["value"] = None
        for key, input in task["inputs"].items():
            if input["property"] is None:
                continue
            prop = input["property"]
            if "." in key and prop["value"] not in [None, {}]:
                root, key = key.split(".", 1)
                root_prop = task["inputs"][root]["property"]
                update_nested_dict(root_prop["value"], key, prop["value"])
                prop["value"] = None


def generate_node_graph(
    pk: int, output: str = None, width: str = "100%", height: str = "600px"
) -> Any:
    """Generate the node graph for the given node pk.
    If in Jupyter, return the graphviz object.
    Otherwise, save the graph to an HTML file.
    """

    from aiida.tools.visualization import Graph
    from aiida import orm
    from IPython.display import IFrame
    import pathlib
    from .svg_to_html import svg_to_html

    in_jupyter = False
    try:
        from IPython import get_ipython

        if get_ipython() is not None:
            in_jupyter = True
    except NameError:
        pass

    graph = Graph()
    calc_node = orm.load_node(pk)
    graph.recurse_ancestors(calc_node, annotate_links="both")
    graph.recurse_descendants(calc_node, annotate_links="both")
    g = graph.graphviz
    if not in_jupyter:
        html_content = svg_to_html(g._repr_image_svg_xml(), width, height)
        if output is None:
            pathlib.Path("html").mkdir(exist_ok=True)
            output = f"html/node_graph_{pk}.html"
        with open(output, "w") as f:
            f.write(html_content)
        return IFrame(output, width=width, height=height)
    return g


def build_task_link(wgdata: Dict[str, Any]) -> None:
    """Create links for tasks.
    Create the links for task inputs using:
    1) workgraph links
    2) if it is a graph builder graph, expose the group inputs and outputs
    sockets.
    """
    # reset task input links
    for name, task in wgdata["tasks"].items():
        for input in task["inputs"]:
            input["links"] = []
        for output in task["outputs"]:
            output["links"] = []
    for link in wgdata["links"]:
        to_socket = [
            socket
            for socket in wgdata["tasks"][link["to_node"]]["inputs"]
            if socket["name"] == link["to_socket"]
        ][0]
        from_socket = [
            socket
            for socket in wgdata["tasks"][link["from_node"]]["outputs"]
            if socket["name"] == link["from_socket"]
        ][0]
        to_socket["links"].append(link)
        from_socket["links"].append(link)


def get_dict_from_builder(builder: Any) -> Dict:
    """Transform builder to pure dict."""
    from aiida.engine.processes.builder import ProcessBuilderNamespace

    if isinstance(builder, ProcessBuilderNamespace):
        return {k: get_dict_from_builder(v) for k, v in builder.items()}
    else:
        return builder


def serialize_workgraph_data(wgdata: Dict[str, Any]) -> Dict[str, Any]:
    from aiida.orm.utils.serialize import serialize

    for name, task in wgdata["tasks"].items():
        wgdata["tasks"][name] = serialize(task)
    wgdata["error_handlers"] = serialize(wgdata["error_handlers"])


def get_workgraph_data(process: Union[int, orm.Node]) -> Optional[Dict[str, Any]]:
    """Get the workgraph data from the process node."""
    from aiida.orm.utils.serialize import deserialize_unsafe
    from aiida.orm import load_node

    if isinstance(process, int):
        process = load_node(process)
    wgdata = process.base.extras.get("_workgraph", None)
    if wgdata is None:
        return
    for name, task in wgdata["tasks"].items():
        wgdata["tasks"][name] = deserialize_unsafe(task)
    wgdata["error_handlers"] = deserialize_unsafe(wgdata["error_handlers"])
    return wgdata


def get_parent_workgraphs(pk: int) -> list:
    """Get the list of parent workgraphs.
    Use aiida incoming links to find the parent workgraphs.
    the parent workgraph is the workgraph that has a link (type CALL_WORK) to the current workgraph.
    """
    from aiida import orm
    from aiida.common.links import LinkType

    node = orm.load_node(pk)
    parent_workgraphs = [[node.process_label, node.pk]]
    links = node.base.links.get_incoming(link_type=LinkType.CALL_WORK).all()
    print(links)
    if len(links) > 0:
        parent_workgraphs.extend(get_parent_workgraphs(links[0].node.pk))
    print(parent_workgraphs)
    return parent_workgraphs


def get_processes_latest(
    pk: int, node_name: str = None, item_type: str = "task"
) -> Dict[str, Dict[str, Union[int, str]]]:
    """Get the latest info of all tasks from the process."""
    import aiida
    from aiida.orm.utils.serialize import deserialize_unsafe
    from aiida.orm import QueryBuilder
    from aiida_workgraph.engine.workgraph import WorkGraphEngine

    tasks = {}
    if item_type == "called_process":
        # fetch the process that called by the workgraph
        node = aiida.orm.load_node(pk)
        for link in node.base.links.get_outgoing().all():
            if isinstance(link.node, aiida.orm.ProcessNode):
                tasks[f"{link.node.process_label}_{link.node.pk}"] = {
                    "pk": link.node.pk,
                    "process_type": link.node.process_type,
                    "state": link.node.process_state.value,
                    "ctime": link.node.ctime,
                    "mtime": link.node.mtime,
                }
    elif item_type == "task":
        node_names = [node_name] if node_name else []
        if node_name:
            projections = [
                f"extras._task_state_{node_name}",
                f"extras._task_process_{node_name}",
            ]
        else:
            projections = []
            process = aiida.orm.load_node(pk)
            node_names = [
                key[12:]
                for key in process.base.extras.keys()
                if key.startswith("_task_state")
            ]
            projections = [f"extras._task_state_{name}" for name in node_names]
            projections.extend([f"extras._task_process_{name}" for name in node_names])
        qb = QueryBuilder()
        qb.append(WorkGraphEngine, filters={"id": pk}, project=projections)
        # print("projections: ", projections)
        results = qb.all()
        # change results to dict
        results = dict(zip(projections, results[0]))
        # print("results: ", results)
        for name in node_names:
            state = results[f"extras._task_state_{name}"]
            task_process = deserialize_unsafe(results[f"extras._task_process_{name}"])
            tasks[name] = {
                "pk": task_process.pk if task_process else None,
                "process_type": task_process.process_type if task_process else "",
                "state": state,
                "ctime": task_process.ctime if task_process else None,
                "mtime": task_process.mtime if task_process else None,
            }

    return tasks


def get_or_create_code(
    computer: str = "localhost",
    code_label: str = "python3",
    code_path: str = None,
    prepend_text: str = "",
):
    """Try to load code, create if not exit."""
    from aiida.orm.nodes.data.code.installed import InstalledCode

    try:
        return orm.load_code(f"{code_label}@{computer}")
    except NotExistent:
        description = f"Code on computer: {computer}"
        computer = orm.load_computer(computer)
        code_path = code_path or code_label
        code = InstalledCode(
            computer=computer,
            label=code_label,
            description=description,
            filepath_executable=code_path,
            default_calc_job_plugin="workgraph.python",
            prepend_text=prepend_text,
        )

        code.store()
        return code


def serialize_properties(wgdata):
    """Serialize the properties.
    Because we use yaml (aiida's serialize) to serialize the data and
    save it to the node.base.extras. yaml can not handle the function
    defined in a scope, e.g., local function in another function.
    So, if a function is used as input, we needt to serialize the function.
    """
    from aiida_workgraph.orm.function_data import PickledLocalFunction
    from aiida_workgraph.tasks.pythonjob import PythonJob
    import inspect

    for _, task in wgdata["tasks"].items():
        if task["metadata"]["node_type"].upper() == "PYTHONJOB":
            PythonJob.serialize_pythonjob_data(task)
        for _, input in task["inputs"].items():
            if input["property"] is None:
                continue
            prop = input["property"]
            if inspect.isfunction(prop["value"]):
                prop["value"] = PickledLocalFunction(prop["value"]).store()


def generate_bash_to_create_python_env(
    name: str,
    pip: list = None,
    conda: dict = None,
    modules: list = None,
    python_version: str = None,
    variables: dict = None,
    shell: str = "posix",
):
    """
    Generates a bash script for creating or updating a Python environment on a remote computer.
    If python_version is None, it uses the Python version from the local environment.
    Conda is a dictionary that can include 'channels' and 'dependencies'.
    """
    import sys

    pip = pip or []
    conda_channels = conda.get("channels", []) if conda else []
    conda_dependencies = conda.get("dependencies", []) if conda else []
    # Determine the Python version from the local environment if not provided
    local_python_version = f"{sys.version_info.major}.{sys.version_info.minor}"
    desired_python_version = (
        python_version if python_version is not None else local_python_version
    )

    # Start of the script
    script = "#!/bin/bash\n\n"

    # Load modules if provided
    if modules:
        script += "# Load specified system modules\n"
        for module in modules:
            script += f"module load {module}\n"

    # Conda shell hook initialization for proper conda activation
    script += "# Initialize Conda for this shell\n"
    script += f'eval "$(conda shell.{shell} hook)"\n'

    script += "# Setup the Python environment\n"
    script += "if ! conda info --envs | grep -q ^{name}$; then\n"
    script += "    # Environment does not exist, create it\n"
    if conda_dependencies:
        dependencies_string = " ".join(conda_dependencies)
        script += f"    conda create -y -n {name} python={desired_python_version} {dependencies_string}\n"
    else:
        script += f"    conda create -y -n {name} python={desired_python_version}\n"
    script += "fi\n"
    if conda_channels:
        script += "EXISTING_CHANNELS=$(conda config --show channels)\n"
        script += "for CHANNEL in " + " ".join(conda_channels) + ";\n"
        script += "do\n"
        script += '    if ! echo "$EXISTING_CHANNELS" | grep -q $CHANNEL; then\n'
        script += "        conda config --prepend channels $CHANNEL\n"
        script += "    fi\n"
        script += "done\n"
    script += f"conda activate {name}\n"

    # Install pip packages
    if pip:
        script += f"pip install {' '.join(pip)}\n"

    # Set environment variables
    if variables:
        for var, value in variables.items():
            script += f"export {var}='{value}'\n"

    # End of the script
    script += "echo 'Environment setup is complete.'\n"

    return script


def create_conda_env(
    computer: Union[str, orm.Computer],
    name: str,
    pip: list = None,
    conda: list = None,
    modules: list = None,
    python_version: str = None,
    variables: dict = None,
    shell: str = "posix",
) -> tuple:
    """Test that there is no unexpected output from the connection."""
    # Execute a command that should not return any error, except ``NotImplementedError``
    # since not all transport plugins implement remote command execution.
    from aiida.common.exceptions import NotExistent
    from aiida import orm

    user = orm.User.collection.get_default()
    if isinstance(computer, str):
        computer = orm.load_computer(computer)
    try:
        authinfo = computer.get_authinfo(user)
    except NotExistent:
        raise f"Computer<{computer.label}> is not yet configured for user<{user.email}>"

    scheduler = authinfo.computer.get_scheduler()
    transport = authinfo.get_transport()

    script = generate_bash_to_create_python_env(
        name, pip, conda, modules, python_version, variables, shell
    )
    with transport:
        scheduler.set_transport(transport)
        try:
            retval, stdout, stderr = transport.exec_command_wait(script)
        except NotImplementedError:
            return (
                True,
                f"Skipped, remote command execution is not implemented for the "
                f"`{computer.transport_type}` transport plugin",
            )

        if retval != 0:
            return (
                False,
                f"The command `echo -n` returned a non-zero return code ({retval})",
            )

        template = """
We detected an error while creating the environemnt on the remote computer, as shown between the bars
=============================================================================================
{}
=============================================================================================
Please check!
    """
        if stderr:
            return False, template.format(stderr)

        if stdout:
            # the last line is the echo 'Environment setup is complete.'
            if not stdout.strip().endswith("Environment setup is complete."):
                return False, template.format(stdout)
            else:
                return True, "Environment setup is complete."

    return True, None


def create_and_pause_process(
    runner: Runner = None,
    process_class: Callable = None,
    inputs: dict = None,
    state_msg: str = "",
) -> Process:
    from aiida.engine.utils import instantiate_process

    process_inited = instantiate_process(runner, process_class, **inputs)
    process_inited.pause(msg=state_msg)
    process_inited.runner.persister.save_checkpoint(process_inited)
    process_inited.close()
    runner.controller.continue_process(process_inited.pid, nowait=True, no_reply=True)
    return process_inited


def recursive_to_dict(attr_dict):
    """
    Recursively convert an AttributeDict to a standard dictionary.
    """
    from aiida.common import AttributeDict
    from plumpy.utils import AttributesFrozendict

    if isinstance(attr_dict, (AttributesFrozendict, AttributeDict)):
        return {k: recursive_to_dict(v) for k, v in attr_dict.items()}
    else:
        return attr_dict


def get_raw_value(identifier, value: Any) -> Any:
    """Get the raw value from a Data node."""
    if identifier in [
        "workgraph.int",
        "workgraph.float",
        "workgraph.string",
        "workgraph.bool",
        "workgraph.aiida_int",
        "workgraph.aiida_float",
        "workgraph.aiida_string",
        "workgraph.aiida_bool",
    ]:
        if value is not None and isinstance(value, orm.Data):
            return value.value
        else:
            return value
    elif (
        identifier == "workgraph.aiida_structure"
        and value is not None
        and isinstance(value, orm.StructureData)
    ):
        content = value.backend_entity.attributes
        content["node_type"] = value.node_type
        return content
    elif isinstance(value, orm.Data):
        content = value.backend_entity.attributes
        content["node_type"] = value.node_type
        return content


def process_properties(task: Dict) -> Dict:
    """Extract raw values."""
    result = {}
    for name, prop in task["properties"].items():
        identifier = prop["identifier"]
        value = prop.get("value")
        result[name] = get_raw_value(identifier, value)
    #
    for name, input in task["inputs"].items():
        if input["property"] is not None:
            prop = input["property"]
            identifier = prop["identifier"]
            value = prop.get("value")
            result[name] = get_raw_value(identifier, value)

    return result


def workgraph_to_short_json(
    wgdata: Dict[str, Union[str, List, Dict]]
) -> Dict[str, Union[str, Dict]]:
    """Export a workgraph to a rete js editor data."""
    wgdata_short = {
        "name": wgdata["name"],
        "uuid": wgdata["uuid"],
        "state": wgdata["state"],
        "nodes": {},
        "links": wgdata["links"],
    }
    #
    for name, task in wgdata["tasks"].items():
        # Add required inputs to nodes
        inputs = [{"name": name} for name in task["inputs"] if name in task["args"]]
        properties = process_properties(task)
        wgdata_short["nodes"][name] = {
            "label": task["name"],
            "node_type": task["metadata"]["node_type"].upper(),
            "inputs": inputs,
            "properties": properties,
            "outputs": [],
            "position": task["position"],
            "children": task["children"],
        }
    # Add links to nodes
    for link in wgdata["links"]:
        wgdata_short["nodes"][link["to_node"]]["inputs"].append(
            {
                "name": link["to_socket"],
            }
        )
        wgdata_short["nodes"][link["from_node"]]["outputs"].append(
            {
                "name": link["from_socket"],
            }
        )
    return wgdata_short



html_template = """
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Rete.js with React in Vanilla JS</title>
    <!-- Import React, ReactDOM, and Babel from CDN -->
    <script src="https://unpkg.com/react@18.2.0/umd/react.development.js" crossorigin></script>
    <script src="https://unpkg.com/react-dom@18.2.0/umd/react-dom.development.js" crossorigin></script>
    <script src="https://unpkg.com/@babel/standalone/babel.min.js"></script>
    <script src="https://cdnjs.cloudflare.com/ajax/libs/react-is/18.2.0/umd/react-is.production.min.js"></script>
    <script src="https://cdn.jsdelivr.net/npm/styled-components@5.3.6/dist/styled-components.min.js"></script>
    <script src="https://unpkg.com/elkjs@0.8.2/lib/elk.bundled.js"></script>

    <!-- Import Rete.js and its plugins from CDN -->
    <script src="https://cdn.jsdelivr.net/npm/rete@2.0.3/rete.min.js"></script>
    <script src="https://cdn.jsdelivr.net/npm/rete-area-plugin@2.0.3/rete-area-plugin.min.js"></script>
    <script src="https://cdn.jsdelivr.net/npm/rete-connection-plugin@2.0.2/rete-connection-plugin.min.js"></script>
    <script src="https://cdn.jsdelivr.net/npm/rete-render-utils@2.0.2/rete-render-utils.min.js"></script>
    <script src="https://cdn.jsdelivr.net/npm/rete-react-plugin@2.0.5/rete-react-plugin.min.js"></script>
    <script src="https://cdn.jsdelivr.net/npm/rete-auto-arrange-plugin@2.0.1/rete-auto-arrange-plugin.min.js"></script>
    <script src="https://cdn.jsdelivr.net/npm/rete-minimap-plugin@2.0.1/rete-minimap-plugin.min.js"></script>
    <script src="https://cdn.jsdelivr.net/npm/rete-scopes-plugin@2.1.0/rete-scopes-plugin.min.js"></script>

    <style>
        .App {
            font-family: sans-serif;
            background: rgb(200, 190, 190);
        }
        .rete {
          position: relative;
          font-size: 1rem;
          margin: 1em;
          border-radius: 1em;
          text-align: left;
        }
        #fullscreen-btn {
            margin-left: 10px;
        }
        body {
            overflow: hidden;
            margin: 0;
            padding: 0;
        }
    </style>
</head>
<body>
    <div id="root"></div>
    <script type="text/babel">

        const { useState, useRef, useEffect } = React;
        const { createRoot } = ReactDOM;
        const { NodeEditor, ClassicPreset } = Rete;
        const { AreaPlugin, AreaExtensions } = ReteAreaPlugin;
        const { ScopesPlugin, Presets: ScopesPresets } = ReteScopesPlugin;
        const { ConnectionPlugin, Presets: ConnectionPresets } = ReteConnectionPlugin;
        const { ReactPlugin, Presets } = ReteReactPlugin;
        const { AutoArrangePlugin, Presets: ArrangePresets, ArrangeAppliers} = ReteAutoArrangePlugin;
        const { MinimapExtra, MinimapPlugin } = ReteMinimapPlugin;
        const { RenderUtils } = ReteRenderUtils;
        const styled = window.styled;

        const workgraphData = __WORKGRAPH_DATA__

        // Define Schemes to use in vanilla JS
        const Schemes = {
            Node: ClassicPreset.Node,
            Connection: ClassicPreset.Connection
        };

        class Node extends ClassicPreset.Node {
          width = 180;
          height = 100;
        }
        class Connection extends ClassicPreset.Connection {}

        function createDynamicNode(nodeData) {
          const node = new Node(nodeData.label);
          // resize the node based on the max length of the input/output names
          let maxSocketNameLength = 0;
          nodeData.inputs.forEach((input) => {
            let socket = new ClassicPreset.Socket(input.name);
            if (!node.inputs.hasOwnProperty(input.name)) {
              node.addInput(input.name, new ClassicPreset.Input(socket, input.name));
              maxSocketNameLength = Math.max(maxSocketNameLength, input.name.length);
            }
          });

          nodeData.outputs.forEach((output) => {
            let socket = new ClassicPreset.Socket(output.name);
            if (!node.outputs.hasOwnProperty(output.name)) {
              node.addOutput(output.name, new ClassicPreset.Output(socket, output.name));
              maxSocketNameLength = Math.max(maxSocketNameLength, output.name.length);
            }
          });
          node.height = Math.max(140, node.height + (nodeData.inputs.length + nodeData.outputs.length) * 35)
          node.width += maxSocketNameLength * 5;

          return node;
        }


        async function addNode(editor, area, nodeData) {
          console.log("Adding node", nodeData);
          const node = createDynamicNode(nodeData);
          await editor.addNode(node);
          editor.nodeMap[nodeData.label] = node; // Assuming each nodeData has a unique ID
          await area.translate(node.id, { x: nodeData.position[0], y: nodeData.position[1] });
        }

        async function addLink(editor, area, layout, linkData) {
          const fromNode = editor.nodeMap[linkData.from_node];
          const toNode = editor.nodeMap[linkData.to_node];
          console.log("fromNode", fromNode, "toNode", toNode);
          let socket;
          if (fromNode && toNode) {
            socket = new ClassicPreset.Socket(linkData.from_socket);
            if (!fromNode.outputs.hasOwnProperty(linkData.from_socket)) {
              fromNode.addOutput(linkData.from_socket, new ClassicPreset.Output(socket, linkData.from_socket));
              fromNode.height += 25; // Increase height of node for each output
              area.update('node', fromNode.id);
            }
            socket = new ClassicPreset.Socket(linkData.to_socket);
            if (!toNode.inputs.hasOwnProperty(linkData.to_socket)) {
              toNode.addInput(linkData.to_socket, new ClassicPreset.Input(socket, linkData.to_socket));
              toNode.height += 25; // Increase height of node for each input
              area.update('node', toNode.id);
            }
            await editor.addConnection(new Connection(fromNode, linkData.from_socket, toNode, linkData.to_socket));
            // await layout(true);

          }
        }

        async function loadJSON(editor, area, layout, workgraphData) {
          for (const nodeId in workgraphData.nodes) {
            const nodeData = workgraphData.nodes[nodeId];
            await addNode(editor, area, nodeData);
          }

          // Adding connections based on workgraphData
          workgraphData.links.forEach(async (link) => { // Specify the type of link here
            await addLink(editor, area, layout, link);
          });

          // Add while zones
          console.log("Adding while zone: ");
          for (const nodeId in workgraphData.nodes) {
            const nodeData = workgraphData.nodes[nodeId];
            const node_type = nodeData['node_type'];
            if (node_type === "WHILE" || node_type === "IF" || node_type === "ZONE") {
              // find the node
              const node = editor.nodeMap[nodeData.label];
              const children = nodeData['children'];
              // find the id of all nodes in the editor that has a label in while_zone
              for (const nodeId in children) {
                const node1 = editor.nodeMap[children[nodeId]];
                node1.parent = node.id;
              }
            }
          }
        }

        /**
        * Defines custom padding for a scope layout.
        * The padding values are used by the ScopesPlugin to avoid node overlapping with the socket of the parent node.
        */
        const customScopePadding = () => ({
          top: 80,
          left: 30,
          right: 30,
          bottom: 50
        });


        async function createEditor(container) {
            const socket = new ClassicPreset.Socket("socket");

            const editor = new NodeEditor(Schemes);
            const area = new AreaPlugin(container);
            const connection = new ConnectionPlugin();
            const render = new ReactPlugin({ createRoot });
            const scopes = new ScopesPlugin({padding: customScopePadding});
            const arrange = new AutoArrangePlugin();

            const minimap = new MinimapPlugin({
              boundViewport: true
            });

            AreaExtensions.selectableNodes(area, AreaExtensions.selector(), {
                accumulating: AreaExtensions.accumulateOnCtrl(),
            });

            render.addPreset(Presets.classic.setup());
            render.addPreset(Presets.minimap.setup({ size: 200 }));

            connection.addPreset(ConnectionPresets.classic.setup());
            scopes.addPreset(ScopesPresets.classic.setup());

            const applier = new ArrangeAppliers.TransitionApplier({
              duration: 500,
              timingFunction: (t) => t,
              async onTick() {
                await AreaExtensions.zoomAt(area, editor.getNodes());
              }
            });

            arrange.addPreset(ArrangePresets.classic.setup());


            editor.use(area);
            // area.use(connection);
            area.use(render);
            area.use(scopes);
            area.use(arrange);
            area.use(minimap);

            async function layout(animate) {
              await arrange.layout({ applier: animate ? applier : undefined });
              AreaExtensions.zoomAt(area, editor.getNodes());
            }

            // Adding nodes based on workgraphData
            const nodeMap = {}; // To keep track of created nodes for linking
            editor.nodeMap = nodeMap;


            return {
              editor: editor,
              area: area,
              layout: layout,
              destroy: () => area.destroy()
            };
        }

        function toggleFullScreen() {
            if (!document.fullscreenElement) {
                document.documentElement.requestFullscreen();
            } else if (document.exitFullscreen) {
                document.exitFullscreen();
            }
        }

        function App() {
            const [editor, setEditor] = useState(null);
            const containerRef = useRef(null);

            useEffect(() => {
                if (containerRef.current && !editor) {
                    createEditor(containerRef.current).then((editor) => {
                        setEditor(editor);
                        loadJSON(editor.editor, editor.area, editor.layout, workgraphData).then(() => {
                          // aplly layout twice to ensure all nodes are arranged
                          editor?.layout(false).then(() => editor?.layout(true));
                        });
                        window.editor = editor;
                    });
                }
                if (document.getElementById('fullscreen-btn')) {
                    document.getElementById('fullscreen-btn').addEventListener('click', toggleFullScreen);
                }
                return () => {
                    if (editor) {
                        editor.destroy();
                    }
                };
            }, [containerRef, editor]);

            return (

                <div className="App">
                    <div>
                      <button onClick={() => editor?.layout(true)}>Arrange</button>
                      <button id="fullscreen-btn">Fullscreen</button>
                    </div>
                    <div ref={containerRef} className="rete" style={{ height: "100vh", width: "100%" }}></div>
                </div>
            );
        }

        const rootElement = document.getElementById("root");
        const root = createRoot(rootElement);

        root.render(
                <App />
        );
    </script>
</body>
</html>
"""



WIDGET_INSTALLATION_MESSAGE = (
    "Widget dependency not found. To visualize the workgraph, please install the widget dependency. "
    "Use 'pip install aiida-workgraph[widget]' if installing from PyPI. "
    "For local source installations, use 'pip install .[widget]' and then build the JavaScript library. "
    "Refer to the documentation for more details."
)



from typing import Optional, Any
from aiida_workgraph.socket import TaskSocket
from node_graph.serializer import SerializeJson, SerializePickle


class SocketAny(TaskSocket, SerializePickle):
    """Any socket."""

    identifier: str = "workgraph.any"

    def __init__(
        self, name, node=None, type="INPUT", index=0, uuid=None, **kwargs
    ) -> None:
        super().__init__(name, node, type, index, uuid=uuid)
        self.add_property("workgraph.any", name, **kwargs)


class SocketNamespace(TaskSocket, SerializePickle):
    """Namespace socket."""

    identifier: str = "workgraph.namespace"

    def __init__(
        self,
        name: str,
        node: Optional[Any] = None,
        type: str = "INPUT",
        index: int = 0,
        uuid: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        super().__init__(name, node, type, index, uuid=uuid)
        # Set the default value to an empty dictionary
        kwargs.setdefault("default", {})
        self.add_property("workgraph.any", name, **kwargs)


class SocketFloat(TaskSocket, SerializeJson):
    """Float socket."""

    identifier: str = "workgraph.float"

    def __init__(
        self, name, node=None, type="INPUT", index=0, uuid=None, **kwargs
    ) -> None:
        super().__init__(name, node, type, index, uuid=uuid)
        self.add_property("workgraph.float", name, **kwargs)


class SocketInt(TaskSocket, SerializeJson):
    """Int socket."""

    identifier: str = "workgraph.int"

    def __init__(
        self, name, node=None, type="INPUT", index=0, uuid=None, **kwargs
    ) -> None:
        super().__init__(name, node, type, index, uuid=uuid)
        self.add_property("workgraph.int", name, **kwargs)


class SocketString(TaskSocket, SerializeJson):
    """String socket."""

    identifier: str = "workgraph.string"

    def __init__(
        self, name, node=None, type="INPUT", index=0, uuid=None, **kwargs
    ) -> None:
        super().__init__(name, node, type, index, uuid=uuid)
        self.add_property("workgraph.string", name, **kwargs)


class SocketBool(TaskSocket, SerializeJson):
    """Bool socket."""

    identifier: str = "workgraph.bool"

    def __init__(
        self, name, node=None, type="INPUT", index=0, uuid=None, **kwargs
    ) -> None:
        super().__init__(name, node, type, index, uuid=uuid)
        self.add_property("workgraph.bool", name, **kwargs)


class SocketAiiDAFloat(TaskSocket, SerializeJson):
    """AiiDAFloat socket."""

    identifier: str = "workgraph.aiida_float"

    def __init__(
        self,
        name: str,
        node: Optional[Any] = None,
        type: str = "INPUT",
        index: int = 0,
        uuid: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        super().__init__(name, node, type, index, uuid=uuid)
        self.add_property("workgraph.aiida_float", name, **kwargs)


class SocketAiiDAInt(TaskSocket, SerializeJson):
    """AiiDAInt socket."""

    identifier: str = "workgraph.aiida_int"

    def __init__(
        self,
        name: str,
        node: Optional[Any] = None,
        type: str = "INPUT",
        index: int = 0,
        uuid: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        super().__init__(name, node, type, index, uuid=uuid)
        self.add_property("workgraph.aiida_int", name, **kwargs)


class SocketAiiDAString(TaskSocket, SerializeJson):
    """AiiDAString socket."""

    identifier: str = "workgraph.aiida_string"

    def __init__(
        self,
        name: str,
        node: Optional[Any] = None,
        type: str = "INPUT",
        index: int = 0,
        uuid: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        super().__init__(name, node, type, index, uuid=uuid)
        self.add_property("workgraph.aiida_string", name, **kwargs)


class SocketAiiDABool(TaskSocket, SerializeJson):
    """AiiDABool socket."""

    identifier: str = "workgraph.aiida_bool"

    def __init__(
        self,
        name: str,
        node: Optional[Any] = None,
        type: str = "INPUT",
        index: int = 0,
        uuid: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        super().__init__(name, node, type, index, uuid=uuid)
        self.add_property("workgraph.aiida_bool", name, **kwargs)


class SocketAiiDAIntVector(TaskSocket, SerializeJson):
    """Socket with a AiiDAIntVector property."""

    identifier: str = "workgraph.aiida_int_vector"

    def __init__(
        self,
        name: str,
        node: Optional[Any] = None,
        type: str = "INPUT",
        index: int = 0,
        uuid: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        super().__init__(name, node, type, index, uuid=uuid)
        self.add_property("workgraph.aiida_int_vector", name, **kwargs)


class SocketAiiDAFloatVector(TaskSocket, SerializeJson):
    """Socket with a FloatVector property."""

    identifier: str = "workgraph.aiida_float_vector"

    def __init__(
        self,
        name: str,
        node: Optional[Any] = None,
        type: str = "INPUT",
        index: int = 0,
        uuid: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        super().__init__(name, node, type, index, uuid=uuid)
        self.add_property("workgraph.aiida_float_vector", name, **kwargs)


class SocketStructureData(TaskSocket, SerializePickle):
    """Any socket."""

    identifier: str = "workgraph.aiida_structuredata"

    def __init__(
        self, name, node=None, type="INPUT", index=0, uuid=None, **kwargs
    ) -> None:
        super().__init__(name, node, type, index, uuid=uuid)
        self.add_property("workgraph.aiida_structuredata", name, **kwargs)



from typing import Dict
from aiida_workgraph import task
from aiida.orm import StructureData, UpfData


@task(
    inputs=[
        {"identifier": "workgraph.string", "name": "pseudo_family"},
        {"identifier": StructureData, "name": "structure"},
    ],
    outputs=[{"identifier": UpfData, "name": "Pseudo"}],
)
def get_pseudo_from_structure(
    pseudo_family: str, structure: StructureData
) -> Dict[str, UpfData]:
    """for input_namespace"""
    from aiida.orm import Group, QueryBuilder

    pseudo_group = (
        QueryBuilder().append(Group, filters={"label": pseudo_family}).one()[0]
    )
    elements = [kind.name for kind in structure.kinds]
    pseudos = {}
    for ele in elements:
        for n in pseudo_group.nodes:
            if ele == n.element:
                pseudos[ele] = n
    return {"Pseudo": pseudos}



from typing import Optional, Dict, Any
import importlib.metadata
import pathlib

import anywidget
import traitlets
from .utils import wait_to_link

try:
    __version__ = importlib.metadata.version("widget")
except importlib.metadata.PackageNotFoundError:
    __version__ = "unknown"

default_value = {"summary": {}, "nodes": {}, "links": [], "logs": [], "pk": []}
default_style = {"width": "90%", "height": "600px"}
default_settings = {"minimap": True}


class NodeGraphWidget(anywidget.AnyWidget):
    _esm = pathlib.Path(__file__).parent / "static" / "widget.js"
    _css = pathlib.Path(__file__).parent / "static" / "widget.css"
    value = traitlets.Dict(default_value).tag(sync=True)
    settings = traitlets.Dict(default_settings).tag(sync=True)
    style = traitlets.Dict(default_style).tag(sync=True)
    states = traitlets.Dict({}).tag(sync=True)
    positions = traitlets.Dict({}).tag(sync=True)

    def __init__(self, parent: Optional[Any] = None, **kwargs: Any) -> None:
        self.parent = parent
        super().__init__(**kwargs)

    def from_workgraph(self, workgraph: Any) -> None:
        from aiida_workgraph.utils import workgraph_to_short_json

        wgdata = workgraph.to_dict()
        wait_to_link(wgdata)
        wgdata = workgraph_to_short_json(wgdata)
        self.value = wgdata

    def from_node(self, node: Any) -> None:
        tdata = node.to_dict()
        tdata.pop("properties", None)
        tdata.pop("executor", None)
        tdata.pop("node_class", None)
        tdata.pop("process", None)
        tdata["label"] = tdata["identifier"]
        for input in tdata["inputs"].values():
            input.pop("property")
        tdata["inputs"] = list(tdata["inputs"].values())
        tdata["outputs"] = list(tdata["outputs"].values())
        wgdata = {"name": node.name, "nodes": {node.name: tdata}, "links": []}
        self.value = wgdata

    def to_html(self, output: str = None, width: str = "100%", height: str = "600px"):
        """Write a standalone html file to visualize the workgraph."""
        from IPython.display import IFrame
        from .html_template import html_template
        import json

        if output is None:
            # create "html" folder if it does not exist
            pathlib.Path("html").mkdir(exist_ok=True)
            output = f"html/{self.value['name']}.html"
        # Replace the placeholder with the actual workgraphData
        html_content = html_template.replace(
            "__WORKGRAPH_DATA__", json.dumps(self.value)
        )
        with open(output, "w") as f:
            f.write(html_content)
        return IFrame(output, width=width, height=height)

    @traitlets.observe("positions")
    def _observe_positions(self, change: Dict[str, Any]) -> None:
        if not self.parent:
            return
        if change["new"]:
            for name, pos in change["new"].items():
                self.parent.nodes[name].position = pos



from aiida.common.lang import classproperty

from aiida.tools.query.mapping import CalculationProjectionMapper


class WorkGraphQueryBuilder:
    """Utility class to construct a QueryBuilder instance for WorkGraph nodes and project the query set."""

    # This tuple serves to mark compound projections that cannot explicitly be projected in the QueryBuilder, but will
    # have to be manually projected from composing its individual projection constituents
    _compound_projections = ("state",)
    _default_projections = (
        "pk",
        "ctime",
        "process_label",
        "state",
        "process_status",
    )
    _valid_projections = (
        "pk",
        "uuid",
        "ctime",
        "mtime",
        "state",
        "process_state",
        "process_status",
        "exit_status",
        "exit_message",
        "sealed",
        "process_label",
        "label",
        "description",
        "node_type",
        "paused",
        "process_type",
        "job_state",
        "scheduler_state",
        "exception",
    )

    def __init__(self, mapper=None):
        if mapper is None:
            self._mapper = CalculationProjectionMapper(self._valid_projections)
        else:
            self._mapper = mapper

    @property
    def mapper(self):
        return self._mapper

    @classproperty
    def default_projections(self):
        return self._default_projections

    @classproperty
    def valid_projections(self):
        return self._valid_projections

    def get_filters(
        self,
        all_entries=False,
        process_state=None,
        process_label=None,
        paused=False,
        exit_status=None,
        failed=False,
        node_types=None,
    ):
        """Return a set of QueryBuilder filters based on typical command line options.

        :param node_types: A tuple of node classes to filter for (must be sub classes of Calculation).
        :param all_entries: Boolean to negate filtering for process state.
        :param process_state: Filter for this process state attribute.
        :param process_label: Filter for this process label attribute.
        :param paused: Boolean, if True, filter for processes that are paused.
        :param exit_status: Filter for this exit status.
        :param failed: Boolean to filter only failed processes.
        :return: Dictionary of filters suitable for a QueryBuilder.append() call.
        """
        # pylint: disable=too-many-arguments
        from aiida.engine import ProcessState

        exit_status_attribute = self.mapper.get_attribute("exit_status")
        process_label_attribute = self.mapper.get_attribute("process_label")
        process_state_attribute = self.mapper.get_attribute("process_state")
        paused_attribute = self.mapper.get_attribute("paused")

        filters = {}

        if node_types is not None:
            filters["or"] = []
            for node_class in node_types:
                filters["or"].append({"type": node_class.class_node_type})

        if process_state and not all_entries:
            filters[process_state_attribute] = {"in": process_state}

        if process_label is not None:
            if "%" in process_label or "_" in process_label:
                filters[process_label_attribute] = {"like": process_label}
            else:
                filters[process_label_attribute] = process_label

        if paused:
            filters[paused_attribute] = True

        if failed:
            filters[process_state_attribute] = {"==": ProcessState.FINISHED.value}
            filters[exit_status_attribute] = {">": 0}

        if exit_status is not None:
            filters[process_state_attribute] = {"==": ProcessState.FINISHED.value}
            filters[exit_status_attribute] = {"==": exit_status}

        return filters

    def get_query_set(
        self,
        relationships=None,
        filters=None,
        order_by=None,
        past_days=None,
        limit=None,
    ):
        """Return the query set of calculations for the given filters and query parameters.

        :param relationships: A mapping of relationships to join on, e.g. {'with_node': Group} to join on a Group. The
            keys in this dictionary should be the keyword used in the `append` method of the `QueryBuilder` to join the
            entity on that is defined as the value.
        :param filters: Rules to filter query results with.
        :param order_by: Order the query set by this criterion.
        :param past_days: Only include entries from the last past days.
        :param limit: Limit the query set to this number of entries.
        :return: The query set, a list of dictionaries.
        """
        import datetime

        from aiida import orm
        from aiida.common import timezone
        from aiida_workgraph.engine.workgraph import WorkGraphEngine

        # Define the list of projections for the QueryBuilder, which are all valid minus the compound projections
        projected_attributes = [
            self.mapper.get_attribute(projection)
            for projection in self._valid_projections
            if projection not in self._compound_projections
        ]
        unique_projections = list(set(projected_attributes))

        if filters is None:
            filters = {}

        if past_days is not None:
            filters["ctime"] = {
                ">": timezone.now() - datetime.timedelta(days=past_days)
            }

        builder = orm.QueryBuilder()
        builder.append(
            WorkGraphEngine, filters=filters, project=unique_projections, tag="process"
        )

        if relationships is not None:
            for tag, entity in relationships.items():
                builder.append(
                    cls=type(entity), filters={"id": entity.pk}, **{tag: "process"}
                )

        if order_by is not None:
            builder.order_by({"process": order_by})
        else:
            builder.order_by({"process": {"ctime": "desc"}})

        if limit is not None:
            builder.limit(limit)

        return builder.iterdict()

    def get_projected(self, query_set, projections):
        """Project the query set for the given set of projections."""
        header = [self.mapper.get_label(projection) for projection in projections]
        result = [header]

        for query_result in query_set:
            result_row = [
                self.mapper.format(projection, query_result["process"])
                for projection in projections
            ]
            result.append(result_row)

        return result



from typing import List, Dict, Any, Union
from fastapi import APIRouter, HTTPException, Query
from aiida import orm

router = APIRouter()


@router.get("/api/datanode-data")
async def read_datanode_data(
    typeSearch: str = Query(None),
    labelSearch: str = Query(None),
) -> List[Dict[str, Any]]:
    from aiida.orm import QueryBuilder, Data
    from aiida_workgraph.web.backend.app.utils import time_ago

    try:
        builder = QueryBuilder()
        filters = {}

        if typeSearch:
            filters["node_type"] = {"like": f"%{typeSearch}%"}

        if labelSearch:
            filters["label"] = {"like": f"%{labelSearch}%"}

        builder.append(
            Data,
            filters=filters,
            project=["id", "uuid", "ctime", "node_type", "label"],
            tag="data",
        )
        builder.order_by({"data": {"ctime": "desc"}})
        records = builder.all()
        data = [
            {
                "pk": pk,
                "uuid": uuid,
                "ctime": time_ago(ctime),
                "node_type": node_type,
                "label": label,
            }
            for pk, uuid, ctime, node_type, label in records
        ]
        return data
    except KeyError:
        raise HTTPException(status_code=404, detail=f"Data node {id} not found")


@router.get("/api/datanode/{id}")
async def read_data_node_item(id: int) -> Dict[str, Any]:

    try:
        node = orm.load_node(id)
        content = node.backend_entity.attributes
        content["node_type"] = node.node_type
        return content
    except KeyError:
        raise HTTPException(status_code=404, detail=f"Data node {id} not found")


# Route for deleting a datanode item
@router.delete("/api/datanode/delete/{id}")
async def delete_data_node(
    id: int,
    dry_run: bool = False,
) -> Dict[str, Union[bool, str, List[int]]]:
    from aiida.tools import delete_nodes

    try:
        # Perform the delete action here
        deleted_nodes, was_deleted = delete_nodes([id], dry_run=dry_run)
        if was_deleted:
            return {
                "deleted": True,
                "message": f"Deleted data node {id}",
                "deleted_nodes": list(deleted_nodes),
            }
        else:
            message = f"Did not delete data node {id}"
            if dry_run:
                message += " [dry-run]"
            return {
                "deleted": False,
                "message": message,
                "deleted_nodes": list(deleted_nodes),
            }
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))



from node_graph.node import Node as GraphNode
from aiida_workgraph import USE_WIDGET
from aiida_workgraph.properties import property_pool
from aiida_workgraph.sockets import socket_pool

if USE_WIDGET:
    from aiida_workgraph.widget import NodeGraphWidget
from aiida_workgraph.collection import (
    WorkGraphPropertyCollection,
    WorkGraphInputSocketCollection,
    WorkGraphOutputSocketCollection,
)
import aiida
from typing import Any, Dict, Optional, Union, Callable, List, Set, Iterable
from aiida_workgraph.utils.message import WIDGET_INSTALLATION_MESSAGE


class Task(GraphNode):
    """Represent a Task in the AiiDA WorkGraph.

    The class extends from node_graph.node.Node and add new
    attributes to it.
    """

    property_pool = property_pool
    socket_pool = socket_pool
    is_aiida_component = False
    _error_handlers = None

    def __init__(
        self,
        context_mapping: Optional[List[Any]] = None,
        process: Optional[aiida.orm.ProcessNode] = None,
        pk: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Task instance.
        """
        super().__init__(
            property_collection_class=WorkGraphPropertyCollection,
            input_collection_class=WorkGraphInputSocketCollection,
            output_collection_class=WorkGraphOutputSocketCollection,
            **kwargs,
        )
        self.context_mapping = {} if context_mapping is None else context_mapping
        self.waiting_on = TaskCollection(parent=self)
        self.process = process
        self.pk = pk
        if USE_WIDGET:
            self._widget = NodeGraphWidget(
                settings={"minmap": False},
                style={"width": "80%", "height": "600px"},
            )
        else:
            self._widget = None
        self.state = "PLANNED"
        self.action = ""

    def to_dict(self) -> Dict[str, Any]:
        from aiida.orm.utils.serialize import serialize

        tdata = super().to_dict()
        tdata["context_mapping"] = self.context_mapping
        tdata["wait"] = [task.name for task in self.waiting_on]
        tdata["children"] = []
        tdata["execution_count"] = 0
        tdata["parent_task"] = [None]
        tdata["process"] = serialize(self.process) if self.process else serialize(None)
        tdata["metadata"]["pk"] = self.process.pk if self.process else None
        tdata["metadata"]["is_aiida_component"] = self.is_aiida_component
        tdata["error_handlers"] = self.get_error_handlers()

        return tdata

    def set_context(self, context: Dict[str, Any]) -> None:
        """Update the context mappings for this task."""
        # all keys should belong to the outputs.keys()
        remain_keys = set(context.keys()).difference(self.outputs.keys())
        if remain_keys:
            msg = f"Keys {remain_keys} are not in the outputs of this task."
            raise ValueError(msg)
        self.context_mapping.update(context)

    def set_from_protocol(self, *args: Any, **kwargs: Any) -> None:
        """Set the task inputs from protocol data."""
        from aiida_workgraph.utils import get_executor, get_dict_from_builder

        executor = get_executor(self.get_executor())[0]
        builder = executor.get_builder_from_protocol(*args, **kwargs)
        data = get_dict_from_builder(builder)
        self.set(data)

    @classmethod
    def new(
        cls, identifier: Union[str, Callable], name: Optional[str] = None
    ) -> "Task":
        """Create a task from a identifier."""
        from aiida_workgraph.tasks import task_pool

        return super().new(identifier, name=name, node_pool=task_pool)

    @classmethod
    def from_dict(cls, data: Dict[str, Any], task_pool: Optional[Any] = None) -> "Task":
        """Create a task from a dictionary. This method initializes a Node instance with properties and settings
        defined within the provided data dictionary. If task_pool is not specified, the default task_pool from
        aiida_workgraph.tasks is used.

        Args:
            data (Dict[str, Any]): A dictionary containing the task's configuration.
            task_pool (Optional[Any]): A pool of node configurations, defaults to None
            which will use the global task_pool.

        Returns:
            Node: An instance of Node initialized with the provided data."""
        from aiida_workgraph.tasks import task_pool as workgraph_task_pool
        from aiida.orm.utils.serialize import deserialize_unsafe

        if task_pool is None:
            task_pool = workgraph_task_pool
        task = GraphNode.from_dict(data, node_pool=task_pool)
        task.context_mapping = data.get("context_mapping", {})
        task.waiting_on.add(data.get("wait", []))
        process = data.get("process", None)
        if process and isinstance(process, str):
            process = deserialize_unsafe(process)
        task.process = process
        task._error_handlers = data.get("error_handlers", [])

        return task

    def reset(self) -> None:
        self.process = None
        self.state = "PLANNED"

    @property
    def error_handlers(self) -> list:
        return self.get_error_handlers()

    def get_error_handlers(self) -> list:
        """Get the error handler function for this task."""
        from aiida_workgraph.utils import build_callable
        from aiida.engine import ExitCode

        if self._error_handlers is None:
            return {}

        handlers = {}
        if isinstance(self._error_handlers, dict):
            for handler in self._error_handlers.values():
                handler["handler"] = build_callable(handler["handler"])
        elif isinstance(self._error_handlers, list):
            for handler in self._error_handlers:
                handler["handler"] = build_callable(handler["handler"])
                handlers[handler["handler"]["name"]] = handler
        # convert exit code label (str) to status (int)
        for handler in handlers.values():
            exit_codes = []
            for exit_code in handler["exit_codes"]:
                if isinstance(exit_code, int):
                    exit_codes.append(exit_code)
                elif isinstance(exit_code, ExitCode):
                    exit_codes.append(exit_code.status)
                else:
                    raise ValueError(f"Exit code {exit_code} is not a valid exit code.")
            handler["exit_codes"] = exit_codes
        return handlers

    def _repr_mimebundle_(self, *args: Any, **kwargs: Any) -> any:

        if self._widget is None:
            print(WIDGET_INSTALLATION_MESSAGE)
            return
        # if ipywdigets > 8.0.0, use _repr_mimebundle_ instead of _ipython_display_
        self._widget.from_node(self)
        if hasattr(self._widget, "_repr_mimebundle_"):
            return self._widget._repr_mimebundle_(*args, **kwargs)
        else:
            return self._widget._ipython_display_(*args, **kwargs)

    def to_html(self, output: str = None, **kwargs):
        """Write a standalone html file to visualize the task."""
        if self._widget is None:
            print(WIDGET_INSTALLATION_MESSAGE)
            return
        self._widget.from_node(self)
        return self._widget.to_html(output=output, **kwargs)


class TaskCollection:
    def __init__(self, parent: "Task"):
        self._items: Set[str] = set()
        self.parent = parent
        self._top_parent = None

    @property
    def graph(self) -> "WorkGraph":
        """Cache and return the top parent of the collection."""
        if not self._top_parent:
            parent = self.parent
            while getattr(parent, "parent", None):
                parent = parent.parent
            self._top_parent = parent
        return self._top_parent

    @property
    def items(self) -> Set[str]:
        return self._items

    def _normalize_tasks(
        self, tasks: Union[List[Union[str, Task]], str, Task]
    ) -> Iterable[str]:
        """Normalize input to an iterable of task names."""
        if isinstance(tasks, (str, Task)):
            tasks = [tasks]
        task_objects = []
        for task in tasks:
            if isinstance(task, str):
                if task not in self.graph.tasks.keys():
                    raise ValueError(
                        f"Task '{task}' is not in the graph. Available tasks: {self.graph.tasks.keys()}"
                    )
                task_objects.append(self.graph.tasks[task])
            elif isinstance(task, Task):
                task_objects.append(task)
            else:
                raise ValueError(f"Invalid task type: {type(task)}")
        return task_objects

    def add(self, tasks: Union[List[Union[str, Task]], str, Task]) -> None:
        """Add tasks to the collection. Tasks can be a list or a single Task or task name."""
        for task in self._normalize_tasks(tasks):
            self._items.add(task)

    def remove(self, tasks: Union[List[Union[str, Task]], str, Task]) -> None:
        """Remove tasks from the collection. Tasks can be a list or a single Task or task name."""
        for task in self._normalize_tasks(tasks):
            if task not in self._items:
                raise ValueError(f"Task '{task.name}' is not in the collection.")
            self._items.remove(task)

    def clear(self) -> None:
        """Clear all items from the collection."""
        self._items.clear()

    def __contains__(self, item: str) -> bool:
        """Check if a task name is in the collection."""
        return item in self._items

    def __iter__(self):
        """Yield each task name in the collection for iteration."""
        return iter(self._items)

    def __len__(self) -> int:
        return len(self._items)

    def __repr__(self) -> str:
        return f"{self._items}"



from typing import Union, Dict
import time
from aiida.engine import calcfunction
from aiida.orm import Int, Float
from aiida_workgraph import task


@calcfunction
def add(
    x: Union[Int, Float], y: Union[Int, Float], t: Union[Int, Float] = 1.0
) -> Dict[str, Union[Int, Float]]:
    """Add node."""
    time.sleep(t.value)
    return {"sum": x + y}


@calcfunction
def greater(
    x: Union[Int, Float], y: Union[Int, Float], t: Union[Int, Float] = 1.0
) -> Dict[str, bool]:
    """Compare node."""
    time.sleep(t.value)
    return {"result": x > y}


@calcfunction
def sum_diff(
    x: Union[Int, Float], y: Union[Int, Float], t: Union[Int, Float] = 1.0
) -> Dict[str, Union[Int, Float]]:
    """Add node."""
    time.sleep(t.value)
    return {"sum": x + y, "diff": x - y}


@task.pythonjob()
def add_pythonjob(x: int, y: int) -> int:
    return x + y



"""Module with `Node` sub class for work processes."""
from typing import Optional, Tuple

from aiida.common.lang import classproperty

from aiida.orm.nodes.process.workflow.workchain import WorkChainNode

__all__ = ("WorkGraphNode",)


class WorkGraphNode(WorkChainNode):
    """ORM class for all nodes representing the execution of a WorkGraph."""

    WORKGRAPH_STATE_INFO_KEY = "workgraph_state_info"

    @classproperty
    def _updatable_attributes(cls) -> Tuple[str, ...]:  # type: ignore
        # pylint: disable=no-self-argument
        return super()._updatable_attributes + (cls.WORKGRAPH_STATE_INFO_KEY,)

    @property
    def workgraph_state_info(self) -> Optional[str]:
        """
        Return the workgraph state info

        :returns: string representation of the workgraph state info
        """
        return self.base.attributes.get(self.WORKGRAPH_STATE_INFO_KEY, None)

    def set_workgraph_state_info(self, workgraph_state_info: str) -> None:
        """
        Set the workgraph state info

        :param state: string representation of the workgraph state info
        """
        return self.base.attributes.set(
            self.WORKGRAPH_STATE_INFO_KEY, workgraph_state_info
        )



from aiida.engine import WorkChain
from aiida import orm
from aiida.engine.processes.workchains.workchain import WorkChainSpec


def select(condition, true, false):
    """Select the data based on the condition."""
    if condition:
        return true
    return false


class GatherWorkChain(WorkChain):
    @classmethod
    def define(cls, spec: WorkChainSpec) -> None:
        """Define the process specification."""

        super().define(spec)
        spec.input_namespace(
            "datas",
            dynamic=True,
            help=('Dynamic namespace for the datas, "{key}" : {Data}".'),
        )
        spec.outline(
            cls.gather,
        )
        spec.output(
            "result",
            valid_type=orm.List,
            required=True,
            help="A list of the uuid of the outputs.",
        )

    def gather(self) -> None:
        datas = self.inputs.datas.values()
        uuids = [data.uuid for data in datas]
        # uuids = gather(uuids)
        self.out("result", orm.List(uuids).store())



from typing import Any


def task_creation_hook(self, task: Any) -> None:
    """Hook for task creation.

    Args:
        task (Task): a task to be created.
    """
    # send message to the widget
    if self.parent._widget is not None:
        self.parent._widget.send(
            {
                "type": "add_task",
                "data": {"label": task.name, "inputs": [], "outputs": []},
            }
        )


def task_deletion_hook(self, task: Any) -> None:
    """Hook for task deletion.

    Args:
        task (Task): a task to be deleted.
    """
    # remove all links to the task
    link_index = []
    for index, link in enumerate(self.parent.links):
        if link.from_node.name == task.name or link.to_node.name == task.name:
            link_index.append(index)
    del self.parent.links[link_index]
    if self._widget is not None:
        self.parent._widget.send({"type": "delete_node", "data": {"name": task.name}})


def link_creation_hook(self, link: Any) -> None:
    """Hook for link creation.

    Args:
        link (Link): a link to be created.
    """
    if self.parent._widget is not None:
        self.parent._widget.send(
            {
                "type": "add_link",
                "data": {
                    "from_node": link.from_node.name,
                    "from_socket": link.from_socket.name,
                    "to_node": link.to_node.name,
                    "to_socket": link.to_socket.name,
                },
            }
        )


def link_deletion_hook(self, link: Any) -> None:
    """Hook for link deletion.

    Args:
        link (Link): a link to be deleted.
    """
    if self.parent._widget is not None:
        self.parent._widget.send(
            {
                "type": "delete_link",
                "data": {
                    "from_node": link.from_node.name,
                    "from_socket": link.from_socket.name,
                    "to_node": link.to_node.name,
                    "to_socket": link.to_socket.name,
                },
            }
        )
